import os
import numpy as np
import re
import matplotlib.pyplot as plt

from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
from agox.databases.database import Database, export_candidates
from agox.utils.plot import plot_atoms, plot_cell
from ase.data.colors import jmol_colors
from ase.data import covalent_radii
from time import strftime
from timeit import default_timer as dt
from argparse import ArgumentParser
import pickle
from scipy.special import erfinv

import warnings

from agox.models.descriptors import Voronoi
from agox.environments import Environment

from ase.io import read
from ase import Atoms

radii_scaling = 30

class Analysis:
    def __init__(
        self,
        directories=[],
        dE=0.05,
        delta_energy=None,
        labels=None,
        save_data=True,
        force_reload=False,
        histogram_kwargs={},
        target_structure=None,
        descriptor=None,
        use_filtering=False,
    ):
        self.initial_directories = directories
        self.num_directories = 0  # This is ticked up when a directory is read.
        self.directories = []
        self.dE = dE
        self.delta_energy = delta_energy
        self.threshold = 0
        self.labels = []
        self.zero_offset = 10000  # Some energy expressions have realistic structures in the positive range, having 0 being the 'failed' energy causes problems for these

        # Plotting settings:
        self.label_fontsize = 12
        self.tick_sizes = 12
        self.colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]

        # Initial the required lists etc:
        self.best_energy = np.zeros(self.num_directories)  # Save the best energy for each directoy.
        self.best_idx = []  # Saves the best index for each directory
        self.best_structures = []
        self.names = []
        self.order = []
        self.energies_list = []
        self.best_energies_list = []

        # Graph filtering
        self.descriptor = descriptor
        self.use_filtering = use_filtering
        self.target_structure = target_structure
        self.all_structures = []
        self.spectra = []

        # Saving and loading:
        self.save_data = save_data
        self.force_reload = force_reload

        # Database filtering:
        self.excluded_db = []

        if len(histogram_kwargs) == 0:
            self.histogram_kwargs = {
                "min_hist_iteration": 0,
                "max_hist_iteration": 0,
                "operation": "best",
                "threshold": 1000,
            }
        else:
            self.histogram_kwargs = histogram_kwargs

        self.xlabel = "Iteration [#]"

    def add_directory(self, directory, label=None, force_reload=False):
        order, names = [], []
        energies, best_energies, best_structure = None, None, None  # Placeholders

        # Check that directory isn't already in self.directories:
        if directory in self.directories:
            print("Directory: {} - already been loaded so skipping it".format(directory))
            return

        # Information for loading/saving data from this directory:
        cwd = os.getcwd()
        full_path = os.path.join(cwd, directory)
        if full_path[-1] != "/":
            full_path += "/"
        data_path = full_path + "BA_data/"
        file_names = ["order", "names", "energies", "best_energies", "best_structure"]
        if self.use_filtering:
            file_names.append("spectra")

        if os.path.exists(data_path) and not force_reload:
            print("Reloading files from {}".format(data_path))
            loaded_data = []
            for file_name in file_names:
                with open(data_path + file_name + ".pckl", "rb") as f:
                    loaded_data.append(pickle.load(f))
            if self.use_filtering:
                order, names, energies, best_energies, best_structure, spectra = loaded_data
            else:
                order, names, energies, best_energies, best_structure = loaded_data

            structures = []
            
        else:
            # Find database files in this directory:
            dbs = []
            print("Loading files from directory: {}".format(directory))
            for f in os.listdir(os.path.join(cwd, directory)):
                if f[-3:] == ".db":
                    order.append(int(re.findall("\d+", f)[0]))  # noqa
                    names.append(f)
                    if os.stat(os.path.join(cwd, directory, f)).st_size < 5000:
                        self.excluded_db.append(order[-1])

            sort_idx = np.argsort(order).astype(int)
            order = np.array(order)[sort_idx]
            names = np.array(names)[sort_idx]

            # Gather data from databases
            for name in names:
                database = Database(os.path.join(cwd, directory, name))
                dbs.append(database)

            energies, best_energies, best_structure = export_candidates(dbs)

            structures = []
            if self.use_filtering:
                spectra = []
                for db in dbs:
                    cands = db.get_all_structures_data()
                    structures.append([Atoms(symbols=c['type'],positions=c['positions'],cell=c['cell'],pbc=c['pbc']) for c in cands])

            # Save some stuff:
            if self.save_data:
                if not os.path.exists(data_path):
                    os.mkdir(data_path)

                things = [order, names, energies, best_energies, best_structure]
                for thing, file_name in zip(things, file_names):
                    with open(data_path + file_name + ".pckl", "wb") as f:
                        pickle.dump(thing, f)

        # Append information to lists
        self.names.append(names)
        self.order.append(order)
        self.energies_list.append(energies)
        self.best_energies_list.append(best_energies)
        self.best_structures.append(best_structure)

        if self.use_filtering:
            self.spectra.append(spectra)            
        self.all_structures.append(structures)

        # When succesfully loaded the directory can be added to the internal count.
        self.directories.append(directory)
        self.num_directories += 1

        # Added to list of plot-labels:
        if label is None:
            idx = directory.rfind("/")

            if idx == len(directory) - 1:
                idx = directory[0:-1].rfind("/")
                self.labels.append(directory[idx + 1 : -1])
            else:
                self.labels.append(directory[idx + 1 : :])
        else:
            self.labels.append(label)

    def compile_information(self):
        """
        Gather information.
        """
        self.iterations = []
        self.restarts = np.zeros(self.num_directories).astype(int)
        for i in range(self.num_directories):
            # self.iterations[i] = np.max([len(arr) for arr in energies_list[i]])
            self.iterations.append([len(arr) for arr in self.energies_list[i]])
            self.restarts[i] = len(self.energies_list[i])

        self.max_num_restarts = np.max(self.restarts).astype(int)
        self.max_num_iterations = np.max([np.max(iterations) for iterations in self.iterations]).astype(int)

        self.energies = (
            np.zeros((self.num_directories, self.max_num_restarts, self.max_num_iterations)) + self.zero_offset
        )
        self.best_energies = (
            np.zeros((self.num_directories, self.max_num_restarts, self.max_num_iterations)) + self.zero_offset
        )

        failed_to_load = []
        for i in range(self.num_directories):
            for j, dat1, dat2 in zip(
                np.arange(len(self.energies_list[i])),
                self.energies_list[i],
                self.best_energies_list[i],
            ):
                try:
                    self.energies[i, j, 0 : len(dat1)] = dat1
                    self.best_energies[i, j, 0 : len(dat2)] = dat2
                    self.best_energies[i, j, len(dat2) : :] = np.nanmin(dat2)
                except ValueError:
                    failed_to_load.append([i, j])

        self.best_energy = np.nanmin(np.nanmin(self.best_energies, axis=2), axis=1)
        self.global_best_energy = np.nanmin(self.best_energies)
        self.global_best_structure = self.best_structures[np.nanargmin(self.best_energy)]

        self.num_atoms = len(self.global_best_structure)

        # Figure out which iteration in each run produced the best iteration:
        safe_best_energies = self.best_energies.copy()
        safe_best_energies[np.isnan(self.best_energies)] = 10e10
        self.best_indicies = np.argmin(safe_best_energies, axis=-1)

        ### Graphs and spectra ###
        if self.use_filtering:
            if self.descriptor.template is None:
                template = Atoms()
                template.cell = self.global_best_structure.cell
                template.pbc = self.global_best_structure.pbc
            else:
                template = self.descriptor.template

            confinement_cell = template.cell
            confinement_corner = np.array([0, 0, 0])
            environment = Environment(
                template=template,
                symbols=str(self.global_best_structure.symbols),
                confinement_cell=confinement_cell,
                confinement_corner=confinement_corner,
            )        
            self.descriptor.environment = environment

            if self.target_structure is not None:                
                self.target_spectrum = self.descriptor.create_features(self.target_structure)
            else:
                self.target_spectrum = self.descriptor.create_features(self.global_best_structure)

            self.spectra_array = np.zeros((self.num_directories, self.max_num_restarts, self.max_num_iterations))
            failed_to_load = []

            for i in range(self.num_directories):
                for j, dat in zip(
                    np.arange(len(self.spectra[i])),
                    self.spectra[i]
                ):
                    try:
                        self.spectra_array[i, j, 0 : len(dat)] = dat
                    except ValueError:
                        failed_to_load.append([i, j])


#            self.spectra = np.array(self.spectra, dtype=object)

            for i in range(self.num_directories):
                if len(self.spectra[i]) == 0:
                    for j in range(self.restarts[i]):
                        for k in range(self.iterations[i][j]):
                            if self.energies[i][j][k] <= self.global_best_energy + self.delta_energy: 
                                spectrum = self.descriptor.create_features(self.all_structures[i][j][k])
                            else:
                                spectrum=0
                                continue

                            if spectrum == self.target_spectrum: # Spectrum equal to target structure spectrum
                                self.spectra_array[i][j][k] = 1
                                break

            if self.save_data:
                cwd = os.getcwd()
                for i in range(len(self.spectra_array)):
                    directory = self.directories[i]
                    full_path = os.path.join(cwd, directory)
                    if full_path[-1] != '/':
                        full_path += '/'
                    data_path = full_path + 'BA_data/'

                    if not os.path.exists(data_path):
                        os.mkdir(data_path)

                    with open(data_path + 'spectra' + '.pckl', 'wb') as f:
                        pickle.dump(self.spectra_array[i], f)  

    def analyse_directories(self):
        for idx, directory in enumerate(self.initial_directories):
            self.add_directory(directory, force_reload=self.force_reload)

        self.compile_information()

    def calculate_CDF(self):
        if self.delta_energy is None:
            delta_energy = self.dE * self.num_atoms
        else:
            self.dE = self.delta_energy / self.num_atoms
            delta_energy = self.delta_energy

        print("Using dE of {:6.4f} eV pr. atom ({} atoms)".format(self.dE, self.num_atoms))

        best_energies = self.best_energies.copy()
        best_energies[np.isnan(best_energies)] = self.global_best_energy + 1e5

        booled = best_energies <= self.global_best_energy + delta_energy
        if self.use_filtering:
            booled2 = self.spectra_array == 1

            booled = np.logical_and(booled, booled2)

        events_np = np.max(booled, axis=2).astype(int)
        times_np = np.argmax(booled, axis=2)
        # times_np[(times_np == 0) * (events_np == 0)] = self.max_num_iterations - 1
        times_np[(events_np == 0)] = self.max_num_iterations - 1
        times_np[(events_np == 1) * (times_np == 0)] = 1
        self.CDF = self.get_stats(times_np, events_np)

        # Determine when the minimum was found:
        self.succes = booled[:, :, -1]

    def get_stats(self, all_times, all_events):
        CDF = []

        for dir_idx in range(len(all_times)):
            times = all_times[dir_idx]
            events = all_events[dir_idx]

            sort_index = np.argsort(times)
            events_sorted = events[sort_index]
            times_sorted = times[sort_index]

            # N_total = len(events) # All restarts are either dead or censored at the end.
            N_total = self.restarts[dir_idx]
            ntimes = len(np.unique(times))

            n = np.ones(ntimes) * N_total
            d = np.zeros(ntimes)

            # Precalculate some things:
            for i, time in enumerate(np.unique(times_sorted)):
                d[i] = np.sum(events_sorted[times_sorted == time])  # Runs that succeed at 'time'
                n[i + 1 :] -= d[i]  # NUmber of runs yet to succeed is reduced.

            # Caculate KM:
            km = np.ones(ntimes + 1)
            times_axis = np.concatenate([[0], np.unique(times_sorted)])
            for i in range(1, len(km)):
                km[i] = km[i - 1] * (1 - (d[i - 1] / n[i - 1]))

            # Uncertainties: Small numbers are for error avoidance, they dont change anything
            # unless you look at digits you shouldnt be looking at.
            V = 1 / (np.log(km + 0.00000001) ** 2 + 0.000001)
            t = 0

            for i in range(1, len(km)):
                if n[i - 1] - d[i - 1] != 0:
                    t += d[i - 1] / (n[i - 1] * (n[i - 1] - d[i - 1]))
                else:
                    t = 0
                V[i] *= t
            V = np.nan_to_num(V, 0)

            alpha = 0.95
            z = np.sqrt(2) * erfinv(2 * (alpha + (1 - alpha) / 2.0) - 1)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                cu = np.exp(-np.exp(np.log(-np.log(km)) - z * np.sqrt(V)))
                cl = np.exp(-np.exp(np.log(-np.log(km)) + z * np.sqrt(V)))

            CDF.append([times_axis, 1 - km, 1 - cl, 1 - cu])

        return CDF

    def plot_CDF(self, ax, factor=1):
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        for i in range(self.num_directories):
            color = colors[i % len(colors)]
            ax.fill_between(
                self.CDF[i][0],
                self.CDF[i][2] * factor,
                self.CDF[i][3] * factor,
                step="post",
                facecolor=color,
                alpha=0.1,
            )
            ax.step(
                self.CDF[i][0],
                self.CDF[i][1] * factor,
                where="post",
                c=color,
                label=self.labels[i],
            )
            max_iterations = np.max(self.iterations[i])
            if not (self.iterations[i] == max_iterations).all():
                idx = np.argmin(
                    np.abs(np.array(self.CDF[i][0]).reshape(-1, 1) - np.array(self.iterations[i]).reshape(1, -1)),
                    axis=0,
                )
                x = self.CDF[i][0][idx]
                x[x == self.CDF[i][0][-2]] = np.array(self.iterations[i])[x == self.CDF[i][0][-2]]
                ax.plot(x, self.CDF[i][1][idx] * factor, "x", color=color)

        ax.set_xlabel(self.xlabel)
        ax.set_ylabel("Success [%]")
        ax.set_ylim([0, 1])
        ax.set_xlim([0, self.max_num_iterations])
        ax.legend()

    def plot_histogram(self, ax, histogram_kwargs=None):
        # Get settings from dict:
        if histogram_kwargs is None:
            min_hist_iteration = self.histogram_kwargs["min_hist_iteration"]
            max_hist_iteration = self.histogram_kwargs["max_hist_iteration"]
            operation = self.histogram_kwargs["operation"]
            threshold = self.histogram_kwargs["threshold"]
        else:
            min_hist_iteration = histogram_kwargs["min_hist_iteration"]
            max_hist_iteration = histogram_kwargs["max_hist_iteration"]
            operation = histogram_kwargs["operation"]
            threshold = histogram_kwargs["threshold"]

        # Clean-up inputs:
        if not operation in ["best", "all"]:
            print("WARNING: Operation {} not implemented - Defaulting to 'best'")
            operation = "best"

        if max_hist_iteration == 0:
            max_hist_iteration = self.energies.shape[2] - 1

        if threshold != 0:
            threshold = self.global_best_energy + threshold

        # Apply 'operation':
        if (
            operation == "best"
        ):  # Histogram over the energies of the BEST structures pr. restart (so min over iterations dimension effectively)
            data = self.best_energies[:, :, max_hist_iteration - 1]
        elif (
            operation == "all"
        ):  # Histogram over the energies of ALL structures pr. iterations within the iteration limits.
            data = self.energies[:, :, min_hist_iteration:max_hist_iteration]

        # Define histogram range:
        min_val = np.nanmin(data[:, :])
        max_val = np.nanmax(data[:, :][data[:, :] < threshold])

        width = self.dE * self.num_atoms

        # num_buckets = abs(int((max_val - min_val) // width))+2
        num_buckets = (np.ceil((max_val - min_val) / width) + 1).astype(int)
        buckets = [min_val + (i * width) for i in range(num_buckets)]

        if buckets == 0:
            buckets = 1

        ymax = 0
        histograms = []
        lines = []
        for dir_idx in range(self.num_directories):
            vals = data[dir_idx, :]
            hist, bin_edges = np.histogram(vals, bins=buckets, range=(min_val, max_val))

            hist = hist.astype(np.float64)
            hist *= 100 / self.restarts[dir_idx]
            patches = []
            for j in range(0, len(bin_edges) - 1):
                p = Rectangle((bin_edges[j], 0), width, hist[j], fill=True)
                patches.append(p)

            pc = PatchCollection(patches, alpha=0.7, facecolor=self.colors[dir_idx], edgecolor="black")
            ax.add_collection(pc)

            ymax = np.max([ymax, np.max(hist)])

            lines.append(
                plt.plot(
                    [-10, -10],
                    [-10, -5],
                    color=self.colors[dir_idx],
                    label=self.labels[dir_idx],
                    linewidth=2,
                )
            )

            histograms.append((hist, bin_edges))

        ax.set_xlim([min_val - width * 2, max_val + width * 2])
        ax.set_ylim([0, ymax + 5])
        ax.plot([0, 0], [0, 0], color="white")
        ax.set_xlabel("Energy [eV]", fontsize=self.label_fontsize)
        ax.set_ylabel("Count [%]", fontsize=self.label_fontsize)

        ax.legend()
        return histograms, lines

    def plot_energy(self, ax, set_labels=True):
        ax.set_title("Energy")
        for dir_idx in range(self.num_directories):
            color = self.colors[dir_idx % len(self.colors)]

            mean = np.nanmean(self.energies[dir_idx, 0 : self.restarts[dir_idx], :], axis=0)

            ax.plot(
                np.nanmean(self.energies[dir_idx, 0 : self.restarts[dir_idx], :], axis=0),
                "-",
                color=color,
                label=self.labels[dir_idx],
                alpha=0.5,
            )
            ax.plot(
                np.nanmean(self.best_energies[dir_idx, 0 : self.restarts[dir_idx], :], axis=0),
                "-",
                label="_nolegend_",
                color=color,
            )

        ax.legend()
        if set_labels:
            ax.set_xlabel(self.xlabel)
            ax.set_ylabel("Energy [eV]")

    def print_info(self, print_all):
        for num, directory in enumerate(self.directories):
            finished = (self.iterations[num] == np.max(self.iterations[num])).all()
            print("=" * 10 + " " + directory + " " + "=" * 10)
            print(f"Finished: {finished}".format(finished))
            print(f"Number of restarts: {self.restarts[num]}")
            print(f"Number of iterations: {np.max(self.iterations[num])}")
            print(f"Success rate: {np.nanmax(self.CDF[num][1]) * 100:6.3f} %")
            print(f"Best energy: {np.nanmin(self.best_energies[num, :, :]):.5f}")

            if print_all:
                order_list = []
                for restart in range(self.restarts[num]):
                    print(
                        f"File: {self.names[num][restart]} - GM found: {self.succes[num, restart]}"
                        f"- Best Energy: {np.min(self.best_energies[num, restart, :])}"
                    )
                    if self.succes[num, restart]:
                        order_list.append(self.order[num][restart])
                # Print the numbers of files that succeeded:
                # print(str(order_list))

    def get_best_structures(self, directory=None):
        structures = []
        cwd = os.getcwd()
        energies = []

        if directory is None:
            directory = np.arange(len(self.directories))

        for i, directory in enumerate([self.directories[i] for i in directory]):
            for j, name in enumerate(self.names[i]):
                try:
                    database_path = os.path.join(cwd, directory, name)
                    # structures.append(get_best(database_path))
                    structures.append(Database(database_path).get_best_structure())
                    structures[-1].db = os.path.join(directory, name)
                    energy = structures[-1].get_potential_energy()
                    energies.append(energy)
                except Exception as e:
                    pass

        indexs = np.argsort(energies).astype(int)
        structures = [structures[i] for i in indexs]
        energies = np.array(energies)[indexs]

        return structures, energies

    def animate_structure(self, ax):
        structures, energies = self.get_best_structures()

        patch_collection = plot_atoms(ax, structures[0], plot_constraint=True)
        plot_cell(ax, structures[0].cell, collection_kwargs=dict(zorder=2))

        ax.set_title("Best Structure")

        return patch_collection, structures, energies


####################################################################################################################
# Interactive: Functions for matplotlib interactivity
####################################################################################################################


class KeyBoardEvent:
    def __init__(self, fig, atoms_patch_collection, structures, energies, named_axes):
        self.ind = 0

        self.structures = structures
        self.energies = energies
        self.num = len(self.structures)

        # Plot objects:
        self.fig = fig
        self.atoms_patch_collection = atoms_patch_collection

        # Axes:
        self.named_axes = named_axes
        self.struct_ax = self.named_axes[0]
        self.hist_ax = self.named_axes[3]

        self.hist = self.hist_ax is not None

        if self.hist:
            limits = self.hist_ax.get_ylim()
            (self.hist_line,) = self.hist_ax.plot([self.energies[0], self.energies[0]], [0, limits[1]], color="black")

    def update_atoms(self, index):
        self.atoms_patch_collection.remove()
        self.atoms_patch_collection = plot_atoms(self.struct_ax, self.structures[index], plot_constraint=True)

        try:
            energy = "${:8.3f}$".format(self.structures[index].get_potential_energy())
        except:
            energy = "NaN"
        title = "{}\n{}".format(self.structures[index].db, energy)
        self.named_axes[0].set_title(title)

    def update_hist(self, index):
        self.hist_line.set_xdata([self.energies[index], self.energies[index]])

    def press(self, event):
        if event.key == "right":
            self.ind += 1
            i = self.ind % self.num
            self.update_atoms(i)

            if self.hist:
                self.update_hist(i)

            self.fig.canvas.draw_idle()

        if event.key == "left":
            self.ind -= 1
            i = self.ind % self.num
            self.update_atoms(i)

            if self.hist:
                self.update_hist(i)

            self.fig.canvas.draw_idle()


def command_line_analysis():
    # Input arguments:
    parser = ArgumentParser()
    parser.add_argument("-d", "--directories", nargs="+", type=str)  # List - Directories
    parser.add_argument("-dE", "--delta_energy", type=float, default=0.05)
    parser.add_argument("-DE", "--Delta_energy", type=float, default=None)
    parser.add_argument("--save_name", type=str, default=None)

    # Flags
    parser.add_argument("-c", "--plot_cdf", action="store_false")  # Bool - Plot CDF
    parser.add_argument("-s", "--plot_structure", action="store_false")  # Bool - Plot structure
    parser.add_argument("-e", "--plot_energies", action="store_true")  # Bool - Plot energy
    parser.add_argument("-hg", "--plot_histogram", action="store_false")  # Bool - Plot histogram
    parser.add_argument("-p", "--print_all", action="store_true")  # Print everything
    parser.add_argument("-v", "--view_best", action="store_true")
    parser.add_argument("-g", "--save", action="store_true")  # Save the figure?
    parser.add_argument("-sd", "--save_data", action="store_false")
    parser.add_argument("-fr", "--force_reload", action="store_true")

    # Histogram settings:
    parser.add_argument(
        "-hgl", "--histogram_limits", nargs="+", default=[0, 0]
    )  # Histogram limits: min iteration, max iteration
    parser.add_argument("-hgop", "--histogram_operation", type=str, default="best")  # Histogram operations: best, all
    parser.add_argument(
        "-hgth", "--histogram_threshold", type=int, default=1000
    )  # Histogram energy threshold: 0 gives all (negative) energies, positive number gives that over the best found.

    # Animation related things:
    parser.add_argument("-a", "--animate", action="store_false")  # Print everything

    # Graph sorting
    parser.add_argument('-uf', '--use_filtering', action='store_true')
    parser.add_argument('-target', '--target_structure', default=None)
    parser.add_argument('-template', '--template', default=None)
    parser.add_argument('-angle', '--angle_from_central_atom', type=float, default=20.)
    parser.add_argument('-n_points', '--number_of_points', type=int, default=8)

    args = parser.parse_args()
    directories = args.directories

    if args.Delta_energy is None:
        dE = args.delta_energy
        delta_energy = None
    else:
        dE = None
        delta_energy = args.Delta_energy

    target_structure = args.target_structure
    if target_structure is not None:
        target_structure = ('').join(target_structure)
        target_structure = read(target_structure)

    n_points = args.number_of_points
    angle = args.angle_from_central_atom

    template = args.template
    if template is not None:
        template = ('').join(template)
        template = read(template)

    descriptor = Voronoi(None, 
            covalent_bond_scale_factor = 1.3, 
            n_points = n_points, 
            angle_from_central_atom=angle, 
            environment = None, 
            template=template)

    use_filtering = args.use_filtering

    # Plots:
    plot_cdf = args.plot_cdf
    plot_structure = args.plot_structure
    plot_energies = args.plot_energies
    plot_histogram = args.plot_histogram
    num_plots = plot_cdf + plot_structure + plot_energies + plot_histogram
    print_all = args.print_all
    view_best = args.view_best

    histogram_kwargs = {
        "min_hist_iteration": int(args.histogram_limits[0]),
        "max_hist_iteration": int(args.histogram_limits[1]),
        "operation": args.histogram_operation,
        "threshold": args.histogram_threshold,
    }

    # Save/reload
    save_data = args.save_data
    force_reload = args.force_reload

    # Saving?
    save = args.save
    save_name = args.save_name

    animate = args.animate

    # Start
    A = Analysis(
        directories,
        dE=dE,
        save_data=save_data,
        force_reload=force_reload,
        histogram_kwargs=histogram_kwargs,
        delta_energy=delta_energy,
        target_structure=target_structure,
        descriptor=descriptor,
        use_filtering=use_filtering,
    )
    # A = Analysis(directories, dE=dE, save_data=True, force_reload=False)

    A.analyse_directories()
    A.calculate_CDF()

    A.print_info(print_all)

    # Open the best structure in the ASE GUI.
    if view_best:
        A.global_best_structure.edit()

    if num_plots > 0:
        fig, axes = plt.subplots(1, num_plots, figsize=(num_plots * 5, 5))
        if num_plots > 1:
            axes = axes.flatten()
        else:
            axes = [axes]

        ax_idx = 0
        if plot_structure:
            if animate:
                atoms_patch_collection, structures, energies = A.animate_structure(axes[ax_idx])
            else:
                A.plot_structure(axes[ax_idx], A.global_best_structure)

            struct_ax = axes[ax_idx]
            ax_idx += 1
        else:
            struct_ax = None

        if plot_cdf:
            A.plot_CDF(axes[ax_idx])
            cdf_ax = axes[ax_idx]
            ax_idx += 1
        else:
            cdf_ax = None

        if plot_energies:
            A.plot_energy(axes[ax_idx])
            energy_ax = axes[ax_idx]
            ax_idx += 1
        else:
            energy_ax = None

        if plot_histogram:
            A.plot_histogram(axes[ax_idx])
            hist_ax = axes[ax_idx]
            ax_idx += 1
        else:
            hist_ax = None

        if animate:
            if plot_structure:
                named_axes = [struct_ax, cdf_ax, energy_ax, hist_ax]
                callback = KeyBoardEvent(fig, atoms_patch_collection, structures, energies, named_axes=named_axes)
                fig.canvas.mpl_connect('key_press_event', callback.press)
            else:
                print("To use the animation, the structure has to be plotted")

        plt.tight_layout()

        if save:
            if save_name is None:
                save_name = directories[0]
            plt.savefig(
                os.path.join(save_name + strftime("%Y_%m_%d") + ".png"),
                bbox_inches="tight",
            )

        plt.show()

if __name__ == "__main__":
    command_line_analysis()



from agox.candidates.ABC_candidate import CandidateBaseClass
from ase.calculators.singlepoint import SinglePointCalculator as SPC

class StandardCandidate(CandidateBaseClass):

    """
    Standard candidate class.
    """

    @classmethod
    def from_atoms(cls, template, atoms):
        candidate =  cls(template=template, positions=atoms.positions, numbers=atoms.numbers, cell=atoms.cell, pbc=atoms.pbc)        
        if hasattr(atoms, 'calc'):
            if atoms.calc is not None:
                if 'energy' in atoms.calc.results:
                    if 'forces' in atoms.calc.results:
                        candidate.calc = SPC(candidate, energy=atoms.get_potential_energy(apply_constraint=False), 
                            forces=atoms.get_forces(apply_constraint=False))
                    else:
                        candidate.calc = SPC(candidate, energy=atoms.get_potential_energy(apply_constraint=False))
        return candidate



import numpy as np
import ray

from agox.collectors.ABC_collector import CollectorBaseClass
from agox.samplers.fixed_sampler import FixedSampler
from agox.utils.ray import RayPoolUser


def remote_generate(generator, sampler, environment):
    return generator(sampler, environment)


class ParallelTemperingCollector(CollectorBaseClass, RayPoolUser):
    name = "ParallelTemperingCollector"

    """
    Parameters
    ----------
    num_candidates : dict or list
        If dict, it must have the form {0: [], 500: []} where the keys are the iteration numbers
        and the values are the number of candidates to generate for that iteration.
        If a list it must provide the number of candidates for each generator used. 
    """

    def __init__(self, generator_dict, dynamic=True, **kwargs):
        # Get the kwargs that are relevant for RayPoolUser
        RayPoolUser.__init__(self)
        self.dynamic = dynamic

        # Get all the generators from the generator_dict
        unique_generators = []
        for generator in generator_dict.values():
            if generator not in unique_generators:
                unique_generators.append(generator)

        # Figure out which unique generator each generator in the generator_dict corresponds to.
        self.mapping = {}
        for key, generator in generator_dict.items():
            self.mapping[key] = unique_generators.index(generator)

        # Give the rest to the CollectorBaseClass.
        CollectorBaseClass.__init__(self, unique_generators, **kwargs)

        self.generator_keys = []
        for generator in self.generators:
            key = self.pool_add_module(generator)
            self.generator_keys.append(key)

    def make_candidates(self):

        if self.dynamic:
            self.update_generators()
        # We need to build the args and kwargs sent to the actors.

        # If sampler is empty and initialized, then use fallback generator
        if len(self.sampler) == 0 and self.sampler.initialized:
            number_of_candidates = self.get_number_of_candidates()
            fallback_generator = self.get_fallback_generator()
            generator_id = self.pool_add_module(fallback_generator)
            modules = [[generator_id]] * np.sum(number_of_candidates)
        else: 
            if self.get_iteration_counter() == 1 or self.get_iteration_counter() is None:
                random_key = self.mapping['random']
                modules = [[self.generator_keys[random_key]]] * self.sampler.sample_size
            else:
                # This specifies which module each actor of the pool will use.
                modules = []
                for temperature in self.sampler.temperatures:
                    modules += [[self.get_generator_key(temperature)]]
        
        environment_id = ray.put(self.environment)
        fixed_samplers = self.get_dummy_samplers()
        if len(fixed_samplers) == 0:
            sampler_ids = [ray.put(self.sampler)] * self.sampler.sample_size
        else:
            sampler_ids = [ray.put(sampler) for sampler in fixed_samplers]
            
        # The args and kwargs passed to the function - in this case the remote_generate
        # function defined above.
        args = [[sampler_id, environment_id] for sampler_id in sampler_ids]
        # kwargs = [{}] * np.sum(number_of_candidates)
        kwargs = [{} for _ in range(np.sum(self.sampler.sample_size))]

        # Generate in parallel using the pool.
        candidates = self.pool_map(remote_generate, modules, args, kwargs)

        # Flatten the output which is a list of lists.
        flat_candidates = []
        for cand_list in candidates:
            for cand in cand_list:
                flat_candidates.append(cand.copy())

        for i, cand in enumerate(flat_candidates):
            cand.add_meta_information("walker_index", i)

        return flat_candidates

    def get_dummy_samplers(self):
        sample = self.sampler.get_all_members()
        samplers = []
        for member in sample:
            sampler = FixedSampler(member)
            samplers.append(sampler)
        return samplers

    def get_number_of_candidates(self):
        return [self.sampler.sample_size]

    @classmethod
    def from_sampler(cls, sampler, environment, amplitudes, random_generator_kwargs=None, **kwargs):
        from agox.generators import RandomGenerator, RattleGenerator

        # Generators to produce candidates structures
        if random_generator_kwargs is None:
            random_generator_kwargs = {}
        random_generator = RandomGenerator(**environment.get_confinement(), **random_generator_kwargs)

        if not len(amplitudes) == len(sampler.temperatures):
            raise ValueError(
                "The number of amplitudes must match the number of temperatures in the sampler."
            )

        # Dict specificies how many candidates are created with and the dict-keys are iterations.
        generator_dict = {}
        generator_dict["random"] = random_generator

        N_atoms = len(environment.get_missing_indices())

        for i, t in enumerate(sampler.temperatures):
            generator_dict[t] = RattleGenerator(
                **environment.get_confinement(),
                rattle_amplitude=amplitudes[i],
                n_rattle=N_atoms,
            )
            generator_dict[t].add_dynamic_attribute("rattle_amplitude")  # The rattle amplitude is dynamic.

        # Collector creates a number of structures in each iteration.
        return cls(
            generator_dict,
            sampler=sampler,
            environment=environment,
            order=1,
            **kwargs
        )

    def update_generators(self):
        for i, temperature in enumerate(self.sampler.temperatures):
            generator = self.get_generator(temperature)
            acceptance = self.sampler.get_acceptance_rate(i, start=-10) # Acceptance rate over the last 10 iterations.

            if acceptance.size == 0:
                continue
            
            # Rattle update amplitude
            if hasattr(generator, 'rattle_amplitude'):
                if acceptance < 0.5:
                    generator.rattle_amplitude *= 0.99
                else:
                    generator.rattle_amplitude = min(5, generator.rattle_amplitude * 1.01)

    def get_generator(self, temperature):
        return self.generators[self.mapping[temperature]]
    
    def get_generator_key(self, temperature):
        return self.generator_keys[self.mapping[temperature]]


import numpy as np
from agox.generators.replace import ReplaceGenerator
from scipy.spatial.distance import cdist

class CenterOfGeometryGenerator(ReplaceGenerator):

    """
    Perturbs referentially, moving atoms to create a cluster around the center of geometry.

    Parameters
    -----------
    selection_percentages : dict
        A dictionary with the keys 'low' and 'high' that determines the percentage of atoms to move.
    extra_radius_amplitude : float
        The amplitude of the extra radius around the COG.
    extra_radius_params : dict
        A dictionary with the keys 'low' and 'high' that determines the range of the extra radius.
        
    """

    name = 'CenterOfGeometryGenerator'

    def __init__(self, selection_percentages={'low':0.25, 'high':0.5}, 
                 extra_radius_amplitude=1, 
                 extra_radius_params={'low':-0.5, 'high':3}, 
                 **kwargs):
        super().__init__(**kwargs)
        self.selection_percentages = selection_percentages
        self.extra_radius_amplitude = extra_radius_amplitude
        self.extra_radius_params = extra_radius_params

    def get_indices_to_move(self, candidate):
        # Calculate the distance to the COG: 
        cog = candidate.get_center_of_geometry()        
        cluster_indices = candidate.get_optimize_indices()
        cog_distances = cdist(candidate.positions[cluster_indices], cog).flatten()

        # Determine how many atoms to move:
        cluster_size = len(cluster_indices)
        atoms_picked = np.floor(np.random.uniform(**self.selection_percentages) * cluster_size).astype(int)

        # Pick the 'atoms_picked' number of atoms with the furthest distance to the COG. 
        picked_cluster_indices = list(cluster_indices[np.flip(np.argsort(cog_distances))[0:atoms_picked]])

        return picked_cluster_indices

    def get_new_position_center(self, candidate, index, indices_to_move):
        return candidate.get_center_of_geometry()
            
    def get_radius(self, candidate):
        cluster_indices = candidate.get_optimize_indices()
        cog = candidate.get_center_of_geometry()        
        cog_distances = cdist(candidate.positions[cluster_indices], cog)
        return np.mean(cog_distances) + self.extra_radius_amplitude * np.random.uniform(**self.extra_radius_params)

    




from abc import ABC, abstractmethod
from agox.observer import Observer
from agox.writer import Writer, agox_writer

import functools

class PostprocessBaseClass(ABC, Observer, Writer):

    """
    Base class for all postprocessors.

    Postprocessors are used to apply common postprocessing steps to all generated candidates.

    Parameters
    ----------
    gets : dict
        Dictionary of get keys, e.g. {'get_key':'candidates'}. Used to select 
        from which entry in the agox.main.State cache the postprocessor should get candidates.
    sets : dict
        Dictionary of set keys, e.g. {'set_key':'candidates'}. Used to select 
        in which entry in the agox.main.State cache the postprocessor should store the postprocessed candidates.
    """

    def __init__(self, gets={'get_key':'candidates'}, sets={'set_key':'candidates'}, 
        order=3, verbose=True, use_counter=True, prefix='', surname=''):
        Observer.__init__(self, gets=gets, sets=sets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)

        self.add_observer_method(self.postprocess_candidates,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='AGOX')

    def update(self):
        """
        Used if the postprocessor needs to continously update, e.g. the training of a surrogate potential. 
        """
        pass

    @abstractmethod
    def postprocess(self, candidate):
        """
        Method that actually do the post_processing
        """
        return postprocessed_candidate

    def process_list(self, list_of_candidates):
        """
        This allows all postproccesors to act on a list of candidates serially.
        This function can be overwritten by sub-class to implement parallelism. 
        """
        processed_candidates = []
        for candidate in list_of_candidates:
                processed_candidate = self.postprocess(candidate)
                processed_candidates.append(processed_candidate)
        return processed_candidates

    def immunity_decorator(func):
        @functools.wraps(func)
        def wrapper(self, candidate):
            if candidate is None: 
                return None
            if candidate.get_postprocess_immunity():
                return candidate
            else:
                return func(self, candidate)
        return wrapper
    
    def immunity_decorator_list(func):
        @functools.wraps(func)
        def wrapper(self, candidates):
            non_immune_candidates = []
            immune_candidates = []
            for candidate in candidates:
                if not candidate.get_postprocess_immunity():
                    non_immune_candidates.append(candidate)
                else:
                    immune_candidates.append(candidate)

            if len(non_immune_candidates) > 0:
                return func(self, non_immune_candidates) + immune_candidates
            else:
                return immune_candidates
        return wrapper

    def __add__(self, other):
        return SequencePostprocess(processes=[self, other], order=self.order)

    @agox_writer
    @Observer.observer_method
    def postprocess_candidates(self, state):    
        candidates = state.get_from_cache(self, self.get_key)
        
        if self.do_check():
            candidates = self.process_list(candidates)
            candidates = list(filter(None, candidates))

        # Add data in write mode - so overwrites! 
        state.add_to_cache(self, self.set_key, candidates, mode='w')
        

class SequencePostprocess(PostprocessBaseClass):

    name = 'PostprocessSequence'

    def __init__(self, processes=[], order=None):
        self.processes = processes
        self.order = order

    def postprocess(self, candidate):
        for process in self.processes:
            candidate = process.postprocess(candidate)

        return candidate

    def process_list(self, list_of_candidates):
        for process in self.processes:
            list_of_candidates = process.process_list(list_of_candidates)
        return list_of_candidates

    def __add__(self, other):
        self.processes.append(other)
        return self
    
    def attach(self, main):
        for j, process in enumerate(self.processes):
            process.update_order(process.postprocess_candidates, order=self.order[0]+j*0.1)
            process.attach(main)



"""
A database records the candidates that have been evaluated during a search. 

Further, because some things need to be trained, e.g. a ML potential model, 
databases can be used to store the training data. The databases all implement
a similar observer-pattern as the main AGOX class, so models can be trained 
whenever data is added to the database.
"""

from .ABC_database import DatabaseBaseClass
from agox.databases.database import Database 
from agox.databases.database_concurrent import ConcurrentDatabase

__all__ = ['DatabaseBaseClass', 'Database', 'ConcurrentDatabase']


import numpy as np
from agox.acquisitors.LCB import LowerConfidenceBoundAcquisitor

class PowerLowerConfidenceBoundAcquisitor(LowerConfidenceBoundAcquisitor):

    name = 'PowerLowerConfindenceBoundAcquisitor'

    def __init__(self, power=0, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.power = power

    def acquisition_function(self, E, sigma):
        return E - self.kappa * sigma ** self.power

    def acquisition_force(self, E, F, sigma, sigma_force):
        return F - self.kappa * self.power * sigma**(self.power-1) * sigma_force




import numpy as np
from agox.models.descriptors.ABC_descriptor import DescriptorBaseClass

class TypeDescriptor(DescriptorBaseClass):

    descriptor_type = 'local'
    name = 'TypeDescriptor'

    def __init__(self, environment, **kwargs):
        super().__init__(environment=environment, **kwargs)
        self.unique_numbers = np.unique(environment.get_all_numbers())

    def get_number_of_centers(self, atoms):
        return len(atoms)
    
    def create_features(self, atoms):
        x = np.zeros((len(atoms), len(self.unique_numbers)))
        for atom in atoms:
            x[atom.index, np.where(self.unique_numbers == atom.number)[0][0]] = 1
        return x
    
    def create_feature_gradient(self, atoms):
        n = len(atoms)
        return np.zeros_like(n, n, 3, len(self.unique_numbers))

    @classmethod
    def from_species(cls, species, **kwargs):
        """
        Create a SOAP descriptor from a list of species.

        Parameters
        ----------
        species : list
            List of species.
        **kwargs : dict
            Keyword arguments for the SOAP descriptor.
        """
        from ase import Atoms
        from agox.environments import Environment

        environment = Environment(
            template=Atoms(""),
            symbols="".join(species),
            use_box_constraint=False,
            print_report=False,
        )
        return cls(environment=environment, **kwargs)


from abc import ABC, abstractmethod
from ase import Atoms
import numpy as np

import functools
from copy import deepcopy
from ase.calculators.singlepoint import SinglePointCalculator

from agox.module import Module
from agox.utils.cache import Cache
    
class CandidateBaseClass(ABC, Atoms, Module):

    """
    Base-class for Candidate objects.

    Parameters
    ------------
    template: ase.Atoms
        Atoms object of the template structure. Does not need to be supplied
        if 'template_indices' are given. 
    template_indices: np.array
        Indices of template atoms. 
    kwargs: 
        Everything that can be supplied to an ASE atoms object, specifically 
        cell, positions and numbers of ALL atoms - including template atoms.
    """

    def __init__(self, template=None, template_indices=None, use_cache=True, **kwargs):
        Atoms.__init__(self, **kwargs) # This means all Atoms-related stuff gets set. 
        Module.__init__(self, use_cache=use_cache)
        self.meta_information = self.info
        
        self.use_cache = use_cache
        self._cache = Cache()

        # Template stuff:        
        if template_indices is not None:
            self.template_indices = template_indices.astype(int)
            self.template = self.get_template()
        elif template is not None:
            self.template = template
            self.template_indices = np.arange(len(template))
        else:
            print('You have not supplied a template, using an empty atoms object without PBC and no specified cell.')
            self.template = Atoms(pbc=self.pbc)            
            self.template_indices = np.arange(0)
        
        self.set_pbc(self.template.get_pbc()) # Inherit PBC's from template.
        
        self.postprocess_immunity = False

        # This will happen eventually when work starts on using multiple templates.
        # But the check doesnt work as intended at the moment.
        # if len(template) > 0:            
        #     assert (self.positions[:len(template)] == template.positions).all(), 'Template and positions do not match'


    @classmethod
    def cache(cls, key):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(self, atoms, *args, **kwargs):
                if not self.use_cache:
                    return func(self, atoms, *args, **kwargs)
                
                full_key = self.cache_key + '/' + key
                if isinstance(atoms, CandidateBaseClass):
                    value = atoms.get_from_cache(full_key)
                    if value is None:
                        value = func(self, atoms, *args, **kwargs)
                        if atoms.use_cache:
                            atoms.set_to_cache(full_key, value)
                else:
                    value = func(self, atoms, *args, **kwargs)
                    
                return value
            return wrapper
        return decorator
        
    def get_from_cache(self, key):
        if not self.use_cache:
            return None
        
        identifier, value = self._cache.get(key, (None, None))
        if identifier is not None:
            if self.compare_identity(identifier):
                return value
        else:
            return None

    def set_to_cache(self, key, value):
        self._cache.put(key, (self.get_identifier(), value))

    def compare_identity(self, identifier):
        for a,b in zip(identifier, self.get_identifier()):
            equal = (a == b).all()
            if not equal:
                return equal
        return equal

    def get_identifier(self):
        return (self.get_atomic_numbers(), self.get_positions(), self.get_cell())


    def add_meta_information(self, name, value):
        """
        Adds an entry to the meta_information dictionary

        Parameters
        -----------
        name: str (preferably, but can take anything that indexes a dict)
            Key to the dict
        value:
            Value to be set in the dict. 
        """
        self.meta_information[name] = value

    def add_write_key(self,value):
        """
        Adds a write key to the candidates meta information.

        Parameters
        -----------
        value: str
            key to add to the write information

        """

        try:
            keys=self.get_meta_information("write_keys")
            if isinstance(keys, str):
                keys=list(keys.split(" "))
            if value not in keys: 
                keys.append(value)
        except: 
            keys=[value]
        self.add_meta_information("write_keys",keys)

    
    def print_properties(self,writer,iteration):
        keys=self.get_meta_information("write_keys")
        strings_list=[]
        write_string="Energy: "+str(iteration)+"  "+str(self.get_potential_energy())+"  "
        
        if keys is not None:
            for key in keys:
                addendum="["+key+":"+" "+self.get_meta_information(key)+"] "
                if len(write_string)+len(addendum) > 120:  
                    strings_list.append(write_string)
                    write_string = ""
                
                write_string+=addendum
                #self.writer('Energy {:06d}: {}'.format(self.get_iteration_counter(), candidate.get_potential_energy()), flush=True)
        strings_list.append(write_string)
        if strings_list is not None:
            for string in strings_list:
                writer(string,flush=True)


    def get_meta_information(self, name):
        """
        Get from the meta_information dict. 

        Parameters
        -----------
        name: str
            Key to get with. 
        
        Returns
        --------
        value - any type
            A copy of the wanted entry in the dict or None if it is not set. 
        """
        try:
            return self.meta_information.get(name, None).copy()
        except AttributeError: 
            # This catches for example 'int' that dont have a copy method. 
            # Ints won't change in-place, but it perhaps custom classes will. 
            return self.meta_information.get(name, None)

    def get_meta_information_no_copy(self, name):
        """
        Get from the meta_information dict without copying

        Parameters
        -----------
        name: str
            Key to get with. 
        
        Returns
        --------
        value - any type
            The wanted entry in the dict or None if it is not set. 
        """
        return self.meta_information.get(name, None)

    def has_meta_information(self, name):
        """
        Get from the meta_information dict without copying

        Parameters
        -----------
        name: str
            Key to get with.         

        Returns
        --------
        bool
            True if the 'name' is a key to meta_information. 
        """
        return name in self.meta_information.keys()

    def pop_meta_information(self, name):
        """
        Pop from the meta_information dict. 

        Parameters
        -----------
        name: str
            Key to get with. 
        
        Returns
        --------
        value - any type
            The wanted entry in the dict or None if it is not set. 
        """
        return self.meta_information.pop(name, None)

    def reset_meta_information(self):
        """
        Resets meta_information
        """
        self.meta_information = dict()

    def get_template(self) -> Atoms:
        """
        Get the template atoms object. 

        Returns
        --------
        atoms
            Template as an Atoms object. 
        """
        return Atoms(numbers=self.numbers[self.template_indices], positions=self.positions[self.template_indices], cell=self.cell, pbc=self.pbc)

    def copy(self):
        """
        Return a copy of candidate object. 

        Not sure if template needs to be copied, but will do it to be safe.

        Returns
        --------
        candidate
            A copy of the candidate object. 
        """
        candidate = self.__class__(template=self.template.copy(), cell=self.cell, pbc=self.pbc, info=self.info,
                               celldisp=self._celldisp.copy())
        candidate.meta_information = self.meta_information.copy()
        
        candidate.arrays = {}
        for name, a in self.arrays.items():
            candidate.arrays[name] = a.copy()

        candidate.constraints = deepcopy(self.constraints)
        
        return candidate

    def copy_calculator_to(self, atoms):
        '''
        Copy current calculator and attach to the atoms object
        '''
        if self.calc is not None and 'energy' in self.calc.results:
            if 'forces' in self.calc.results:
                calc = SinglePointCalculator(atoms, energy=self.calc.results['energy'],
                                             forces=self.calc.results['forces'])
            else:
                calc = SinglePointCalculator(atoms, energy=self.calc.results['energy'])
            atoms.set_calculator(calc)

    def set_postprocess_immunity(self, state):
        """
        Set postprocess immunity. 

        Parameters
        -----------
        state: bool
            True if Candidate should ignore postprocessor steps. 
        """
        self.postprocess_immunity = state

    def get_postprocess_immunity(self):
        """
        Get postprocess immunity: 

        Returns
        --------
        bool
            True if immune to postprocessing
        """
        return self.postprocess_immunity

    def get_property(self, key):
        """
        Get property from calculator. 

        Parameters
        -----------
        key
            Key used to index calc.get_property

        Returns
        --------
        value
            Value of calc.get.property(key)
        """

        return self.calc.get_property(key)

    def get_template_indices(self):
        """
        Returns
        --------
        np.array
            Array of template indices
        """
        return self.template_indices

    def get_optimize_indices(self) -> np.array:
        """
        Returns
        --------
        np.array
            Indices of of atoms that are part of the search. 
        """
        return np.arange(len(self.template), len(self))

    def get_center_of_geometry(self, all_atoms=False) -> np.array:
        """
        Returns the center of geometry.

        Parameters
        -----------
        all_atoms: bool
            If True all atoms are included, if False only non-template atoms are. 
        """

        if all_atoms:
            return np.mean(self.positions, axis=0).reshape(1, 3)
        else:
            return np.mean(self.positions[self.get_optimize_indices()], axis=0).reshape(1, 3)

    def set_calculator(self, calc):
        """
        "Old" ASE syntax for setting calculator, avoids the annoying warning and 
        uses this better syntax.

        Parameters
        -----------
        calc: calculator
            Calculator to set.
        """
        self.calc = calc

def disable_cache(candidate):
    # Can be either Candidate or Atoms.
    if isinstance(candidate, CandidateBaseClass):
        candidate.use_cache = False

def switch_cache(candidate, state):
    prev_state = False
    if isinstance(candidate, CandidateBaseClass):
        prev_state = candidate.use_cache
        candidate.use_cache = state
    return prev_state



import numpy as np
from agox.samplers.ABC_sampler import SamplerBaseClass
from agox.candidates.ABC_candidate import CandidateBaseClass
from copy import deepcopy

class FixedSampler(SamplerBaseClass):

    """
    Sampler that always returns the same sample.

    Parameters
    ----------
    sample : list
        List of candidates that will be returned by the sampler.
    p : list
        List of probabilities for each candidate in the sample. 
        If None, all candidates have equal probability.
    """

    name = 'FixedSampler'

    def __init__(self, sample, p=None):

        if isinstance(sample, CandidateBaseClass):
            self.sample = [sample]

        if isinstance(sample, list):
            for s in sample:
                assert isinstance(s, CandidateBaseClass), 'Sample must be a list of Candidate objects'
            self.sample = sample

        if p is None:
            p = np.ones(len(self.sample)) / len(self.sample)
        else:
            assert len(p) == len(self.sample), 'p must have the same length as sample'
            p = np.array(p)

    def setup(self, all_candidates):
        pass



import matplotlib

matplotlib.use("Agg")
from argparse import ArgumentParser

import numpy as np
from ase import Atoms

from agox import AGOX
from agox.databases import Database
from agox.environments import Environment
from agox.evaluators import LocalOptimizationEvaluator
from agox.generators.block_generators import RattleBlockGenerator
from agox.samplers import MetropolisSampler

# Manually set seed and database-index
seed = 42
database_index = 0

# Using argparse if e.g. using array-jobs on Slurm to do several independent searches.
# from argparse import ArgumentParser
# parser = ArgumentParser()
# parser.add_argument('-i', '--run_idx', type=int, default=0)
# args = parser.parse_args()

# seed = args.run_idx
# database_index = args.run_idx

##############################################################################
# Calculator
##############################################################################

from ase.calculators.emt import EMT

calc = EMT()

##############################################################################
# System & general settings:
##############################################################################

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au2Ni2",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

# Database
db_path = "db{}.db".format(database_index)  # From input argument!
database = Database(filename=db_path, order=3)

##############################################################################
# Search Settings:
##############################################################################

sampler = MetropolisSampler(temperature=0.25, order=3)


building_block = Atoms("AuNi", positions=[[0, 0, 0], [2.4, 0, 0]])

rattle_generator = RattleBlockGenerator(
    **environment.get_confinement(),
    building_blocks=[building_block],
    N_blocks=[2],
    environment=environment,
    sampler=sampler,
    order=1,
)

# Wont relax fully with steps:5 - more realistic setting would be 100+.
evaluator = LocalOptimizationEvaluator(
    calc,
    gets={"get_key": "candidates"},
    optimizer_run_kwargs={"fmax": 0.05, "steps": 5},
    store_trajectory=False,
    order=2,
    constraints=environment.get_constraints(),
)

##############################################################################
# Let get the show running!
##############################################################################

agox = AGOX(rattle_generator, database, sampler, evaluator, seed=seed)

agox.run(N_iterations=10)



import numpy as np
from ase.neighborlist import NeighborList
from ase.calculators.calculator import Calculator
from scipy.spatial.distance import cdist
from matscipy.neighbours import neighbour_list as msp_nl

class ComplementaryEnergyGaussianCalculator(Calculator):
    """
    Complementary energy calculator based on Gaussians.

    Parameters
    ------------
    descriptor: Local descriptor
        Used to evaluate energy and forces
    attractors: List of attractors calculated using the same descriptor
        Used to evaluate energy and forces
    mover_indices: List of integers
        Atoms with given indices contribute to the complementary energy
    sigma: float
        Width of Gaussian
    dx: float
        Displacement distance of atoms when calculating forces using
        finite difference method
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, descriptor, attractors = None, mover_indices = None, sigma=10, dx=0.001):
        super().__init__()
        self.descriptor = descriptor
        self.attractors = attractors
        self.mover_indices = mover_indices
        self.sigma = sigma
        self.dx = dx

    def calculate(self, atoms, properties=[], *args, **kwargs):
        self.results['energy'] = self.get_ce_energy(atoms)
        if 'forces' in properties:
            self.results['forces'] = self.get_numerical_forces(atoms)
            
    def get_local_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)
        CE = []
        for a in self.mover_indices:
            attractor_index = np.argmin(cdist(features[a].reshape(1, -1), self.attractors))
            attractor = self.attractors[attractor_index]
            CE.append(np.exp(-np.linalg.norm(features[a]-attractor)**2/(2*self.sigma**2)))

        CE = np.array(CE)
        return CE

    def get_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)
        CE = 0
        for a in self.mover_indices:            
            attractor_index = np.argmin(cdist(features[a].reshape(1, -1), self.attractors))
            attractor = self.attractors[attractor_index]
            CE += -np.exp(-np.linalg.norm(features[a]-attractor)**2/(2*self.sigma**2))

        return CE
    
    def get_numerical_forces(self, atoms):
        F = np.zeros((len(atoms), 3))
        for a in self.mover_indices:
            for d in range(3):                
                atoms.positions[a, d] += self.dx
                ef = self.get_ce_energy(atoms)
                atoms.positions[a, d] -= 2 * self.dx
                em = self.get_ce_energy(atoms)
                atoms.positions[a, d] += self.dx

                F[a, d] = -(ef-em)/(2*self.dx)
        return F

class ComplementaryEnergyKernelMethod(Calculator):
    """
    Complementary energy calculator based on Kernel method.

    Parameters
    ------------
    descriptor: Local descriptor
        Used to evaluate energy and forces
    attractors: List of attractors calculated using the same descriptor
        Used to evaluate energy and forces
    mover_indices: List of integers
        Atoms with given indices contribute to the complementary energy
    sigma: float
        Width of Gaussian
    dx: float
        Displacement distance of atoms when calculating forces using
        finite difference method
    """

    implemented_properties = ['energy', 'forces']

    def __init__(self, descriptor, attractors = None, mover_indices = None, sigma = 10, dx=0.001):
        super().__init__()
        self.descriptor = descriptor
        self.attractors = attractors
        self.mover_indices = mover_indices
        self.sigma = sigma
        self.dx = dx

    def calculate(self, atoms, properties=[], *args, **kwargs):
        self.results['energy'] = self.get_ce_energy(atoms)
        if 'forces' in properties:
            self.results['forces'] = self.get_numerical_forces(atoms)

    def get_local_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)

        CE = []
        for a in self.mover_indices:
            e = -np.exp(-cdist(features[a].reshape(1, -1), self.attractors)**2/(2*self.sigma**2))
            CE.append(np.sum(e))

        CE = np.array(CE)
        return CE

    def get_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)

        CE = 0
        for a in self.mover_indices:           
            e = -np.exp(-cdist(features[a].reshape(1, -1), self.attractors)**2/(2*self.sigma**2))
            CE += np.sum(e)

        return CE

    def get_numerical_forces(self, atoms):
        F = np.zeros((len(atoms), 3))
        for a in self.mover_indices:
            for d in range(3):                
                atoms.positions[a, d] += self.dx
                ef = self.get_ce_energy(atoms)
                atoms.positions[a, d] -= 2 * self.dx
                em = self.get_ce_energy(atoms)
                atoms.positions[a, d] += self.dx

                F[a, d] = -(ef-em)/(2*self.dx)
        return F

class ComplementaryEnergyDistanceCalculator(Calculator):
    """
    Complementary energy calculator based on distances.

    Parameters
    ------------
    descriptor: Local descriptor
        Used to evaluate energy and forces
    attractors: List of attractors calculated using the same descriptor
        Used to evaluate energy and forces
    mover_indices: List of integers
        Atoms with given indices contribute to the complementary energy
    exponent: float
        Exponent of the distance. A value of 1 corresponds to Euclidean distance
    dx: float
        Displacement distance of atoms when calculating forces using
        finite difference method
    """
    
    implemented_properties = ['energy', 'forces']

    def __init__(self, descriptor, attractors = None, mover_indices = None, exponent = 1, dx=0.001):
        super().__init__()
        self.exponent = exponent
        self.descriptor = descriptor
        self.attractors = attractors
        self.mover_indices = mover_indices
        self.dx = dx

    def calculate(self, atoms, properties=[], *args, **kwargs):
        self.results['energy'] = self.get_ce_energy(atoms)
        if 'forces' in properties:
#            self.results['forces'] = self.get_numerical_forces(atoms)
            self.results['forces'] = self.get_forces(atoms)

    def get_local_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)
        CE = []
        for a in range(len(atoms)):
            attractor_index = np.argmin(cdist(features[a].reshape(1, -1), self.attractors))
            attractor = self.attractors[attractor_index]
            CE.append(np.linalg.norm(features[a]-attractor) ** self.exponent)

        CE = np.array(CE)
        return CE

    def get_ce_energy(self, atoms):
        features = self.descriptor.get_features(atoms)
        CE = 0
        for a in self.mover_indices:            
            attractor_index = np.argmin(cdist(features[a].reshape(1, -1), self.attractors))
            attractor = self.attractors[attractor_index]
            CE += np.linalg.norm(features[a]-attractor) ** self.exponent

        return CE

    def get_numerical_forces(self, atoms):
        F = np.zeros((len(atoms), 3))
        for a in self.mover_indices:
            for d in range(3):                
                atoms.positions[a, d] += self.dx
                ef = self.get_ce_energy(atoms)
                atoms.positions[a, d] -= 2 * self.dx
                em = self.get_ce_energy(atoms)
                atoms.positions[a, d] += self.dx

                F[a, d] = -(ef-em)/(2*self.dx)

        return F

    def get_forces(self, atoms):
        F = np.zeros((len(atoms), 3))

        features = self.descriptor.get_features(atoms)
        features_derivatives = self.descriptor.get_feature_gradient(atoms)
        mover_attractor_index = np.argmin(cdist(features, self.attractors), axis = 1)
        feature_dvec = features - self.attractors[mover_attractor_index]
        feature_dists = np.linalg.norm(feature_dvec, axis = 1)

        for i in self.mover_indices:
            if feature_dists[i] == 0:
                continue

            scalar_vector = self.exponent * feature_dvec[i] * feature_dists[i] ** (self.exponent - 2)
            scalar_vector = scalar_vector.reshape(-1) * features_derivatives[i]
            F[i, :] += np.sum(scalar_vector[i], axis = 1).reshape(-1)

            for j in self.mover_indices:
                if j != i:
                    F[j, :] += np.sum(scalar_vector[j], axis = 1).reshape(-1)

        return - F



import ray
from agox.utils.ray.startup import ray_startup
from agox.utils.ray.pool import Pool
from agox.observer import Module
import traceback


class PoolFailedtoStartError(Exception):
    def __init__(self, attempts, message="Failed to start pool after {} attempts."):
        message = message.format(attempts)
        super().__init__(message)


def make_ray_pool(ray_stats, attempt=0, max_attempts=10, **kwargs):
    try:
        # Make the pool:
        ray_pool_instance = Pool(
            num_actors=int(ray_stats["CPU"]))

        print(f"Attempt {attempt}: Pool started successfully.")
        return ray_pool_instance

    except Exception as e:
        print(f"Attempt {attempt}: Failed to correctly start pool.")
        print(f"Error: \n\t{ e}")
        print(f"Traceback: \n{traceback.format_exc()}")

        if attempt < max_attempts:
            return make_ray_pool(
                ray_stats, attempt=attempt + 1, max_attempts=max_attempts, **kwargs
            )
        else:
            raise PoolFailedtoStartError(attempt)


global ray_pool_instance
ray_pool_instance = None

def get_ray_pool(**kwargs):
    # Start Ray - this doesn't do anything if Ray is already started.
    ray_stats = ray_startup(**kwargs)

    # Make the Pool if it doesn't exist already.
    global ray_pool_instance
    if ray_pool_instance is None:
        ray_pool_instance = make_ray_pool(ray_stats, **kwargs)
    return ray_pool_instance, ray_stats

def configure_ray_pool(cpu_count: int = None,
                memory: int = None,
                tmp_dir: str = None,
                include_dashboard: bool = False,
                max_grace_period: float = 15.0, 
                keep_log_files: bool = False):
    """
    Configure the settings for the Ray pool.

    Parameters:
    -----------
    cpu_count: int = None,
    memory: int = None,
    tmp_dir: str = None,
    include_dashboard: bool = False,
    max_grace_period: float = 15.0,
    """
    global ray_pool_instance
    if ray_pool_instance is not None:
        raise ValueError("Pool already started. Cannot reconfigure.")

    pool, _ = get_ray_pool(cpu_count=cpu_count, 
                           memory=memory, 
                           tmp_dir=tmp_dir, 
                           include_dashboard=include_dashboard, 
                           max_grace_period=max_grace_period, 
                           keep_log_files=keep_log_files)
    return None

def reset_ray_pool(**kwargs):
    """
    Should only be used for testing.
    """
    pool, _ = get_ray_pool(**kwargs)
    pool.reset_pool()



import os
import numpy as np
from agox.candidates import StandardCandidate
from ase.io import read, write
import pickle
import pytest
import glob

import sys
from pathlib import Path

if sys.version_info >= (3, 9):
    from importlib.resources import files
else:
    from importlib_resources import files

test_folder_path = os.path.join(files('agox'), 'test/')

test_data_dicts = [
    {'path':'datasets/AgO-dataset.traj', 'remove':6, 'name':'AgO'}, 
    {'path':'datasets/B12-dataset.traj', 'remove':12, 'name':'B12'},        
    {'path':'datasets/C30-dataset.traj', 'remove':30, 'name':'C30'},
    ]

for dictionary in test_data_dicts:
    dictionary['path'] = os.path.join(test_folder_path, dictionary['path'])

class TemporaryFolder:

    def __init__(self, path):
        d = path / ""
        if not os.path.exists(d):
            d.mkdir()
        self.start_dir = os.getcwd()
        self.path = d

    def __enter__(self):
        os.chdir(self.path)

    def __exit__(self, *args):
        os.chdir(self.start_dir)

def compare_candidates(atoms_1, atoms_2, tolerance):

    if atoms_1 is None and atoms_2 is None:
        return True

    np.testing.assert_allclose(atoms_1.positions, atoms_2.positions, **tolerance)
    np.testing.assert_allclose(atoms_1.cell, atoms_2.cell, **tolerance)
    np.testing.assert_allclose(atoms_1.numbers, atoms_2.numbers, **tolerance)
    return True

def get_test_environment(path, remove):
    from agox.environments import Environment

    atoms = read(path, index=0)
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    numbers = atoms.get_atomic_numbers()[len(atoms)-remove:]

    template = atoms.copy()
    del template[len(template)-remove:len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, 
            confinement_corner=corner)

    return environment

def get_test_data(path, environment):
    template = environment.get_template()
    atoms_data = read(path, index=':')
    candidate_data = [StandardCandidate.from_atoms(template, atoms) for atoms in atoms_data]
    return candidate_data

def get_test_sampler(data):
    from agox.samplers import SamplerBaseClass

    class DummySampler(SamplerBaseClass):
        
        name = 'DummySampler'
        initialized = False

        def __init__(self, data):
            self.sample = data

        def setup(self, *args, **kwargs):
            return None

    return DummySampler(data)

def get_name(module_name, subfolder, dataset_name, parameter_index):
    folder  = 'expected_outputs/'
    name = f'{test_folder_path}{subfolder}{folder}{module_name}_data{dataset_name}_parameter{parameter_index}.pckl'
    return name

def save_expected_data(name, data):
    check_file_is_deleted(name)
    with open(name, 'wb') as f:
        pickle.dump(data, f)

def load_expected_data(name):
    with open(name, 'rb') as f:
        data = pickle.load(f)
    return data

def label_dict_list(list_of_dicts):
    for index, dictionary in enumerate(list_of_dicts):
        list_of_dicts[index] = (dictionary, index)
    return list_of_dicts

@pytest.fixture(params=test_data_dicts)
def environment_and_dataset(request):
    from agox.environments import Environment
    from agox.candidates import StandardCandidate
    
    atoms = read(request.param['path'])
    cell = atoms.get_cell()
    corner = np.array([0, 0, 0])
    remove = request.param['remove']
    numbers = atoms.get_atomic_numbers()[len(atoms)-remove:]

    template = read(request.param['path'])
    del template[len(template)-remove:len(template)]
    environment = Environment(template=template, numbers=numbers, confinement_cell=cell, 
            confinement_corner=corner)

    data = read(request.param['path'], ':')
    candidates = [StandardCandidate.from_atoms(template, a) for a in data]

    return environment, candidates

def check_folder_is_empty(path):

    path = Path(path)

    if not path.exists():
        print('Creating path')
        path.mkdir(parents=True)

    # files = glob.glob(str(path) + '/*')
    files = list(path.glob('*'))
    assert len(files) == 0, 'Expected path is not empty, delete the files if you are sure you want to remake them!'

def check_file_is_deleted(filepath):
    assert not os.path.exists(filepath), 'File already exists - manually delete it if youre sure you want to recreate it.'



from agox.main import State

class FakeObserver:

    def __init__(self):
        self.set_values = ['candidates']
        self.get_values = ['candidates']

def test_empty_cache():

    state = State()

    # The state starts with an empty cache.
    assert state.cache == {}

    # The state can add to the cache.
    state.add_to_cache(FakeObserver(), 'candidates', [], 'w')

    # The state can get from the cache.
    assert state.get_from_cache(FakeObserver(), 'candidates') == []







import numpy as np
from agox.generators.complementary_energy.attractor_methods.ABC_attractors import AttractorMethodBaseClass

class AttractorInterpolation(AttractorMethodBaseClass):
    """
    Method for picking attractors from current structure using interpolation.

    Parameters
    -----------
    descriptor: Local descriptor
        Used to calculate local features
    possible_number_of_attractors: int or list with two ints
        If int, then that exact amount of attractors will be created. If a list,
        the number of attractors will be a random number between the two given numbers
    attractors_from_template: Bool
        Whether attractors can be picked as feature vectors of template atoms
    predefined_attractors: List of attractors
        If not None, the specified attractors will be given to the CE generator
    order: int
        When to update the structures by checking the database
    """

    name = 'AttractorInterpolation'

    def __init__(self, descriptor, possible_number_of_attractors = [3, 6], attractors_from_template = True, predefined_attractors = None, order = 5):
        super().__init__(order = order)
        self.descriptor = descriptor
        self.possible_number_of_attractors = possible_number_of_attractors
        self.attractors_from_template = attractors_from_template
        self.predefined_attractors = predefined_attractors

    def get_attractors(self, structure):

        if self.predefined_attractors is not None:
            return self.predefined_attractors
        
        template = structure.get_template()
        n_template = len(template)
        atomic_numbers = structure.get_atomic_numbers()
        types = list(set(atomic_numbers))

        if isinstance(self.possible_number_of_attractors, int):
            number_of_attractors = self.possible_number_of_attractors
        else:
            number_of_attractors = np.random.randint(self.possible_number_of_attractors[0], self.possible_number_of_attractors[1])

        features = self.descriptor.get_features(structure)
        if not self.attractors_from_template:
            features = features[n_template:]
            atomic_numbers = atomic_numbers[n_template:]

        attractor_types = np.random.choice(types, size = number_of_attractors, replace = True)
        attractors = np.zeros((number_of_attractors, len(features[0])))
        for _, attractor_type in enumerate(attractor_types):
            filter = atomic_numbers == attractor_type
            possible_indices = np.array(range(len(features)))[filter]
            index1, index2 = np.random.choice(possible_indices, size = 2, replace = False)
            attractor = 1/2 * (features[index2] - features[index1]) + features[index1]
            attractors[_, :] = attractor

        return attractors


import os
from pathlib import Path

import io
from contextlib import redirect_stdout, redirect_stderr

import numpy as np

import torch
import torchmetrics
import pytorch_lightning as pl
from pytorch_lightning import seed_everything

import schnetpack as spk
import schnetpack.transform as trn
from schnetpack.data import AtomsDataModule, ASEAtomsData
from schnetpack.representation import PaiNN
from schnetpack.interfaces import AtomsConverter

from ase.calculators.singlepoint import SinglePointCalculator as SPC
from ase.calculators.calculator import Calculator, all_changes

from agox.models.ABC_model import ModelBaseClass
from agox.writer import agox_writer
from agox.observer import Observer
from ase import Atoms

import logging
logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)


class SchNetPackModel(ModelBaseClass):
    name = 'SchNetPack-model'

    implemented_properties = ['energy', 'forces']

    dynamic_attributes = ['nnpot']

    """ SchNetPack PaiNN model

    Parameters
    ----------
        
    """

    def __init__(
            self,
            cutoff=6.,
            dataset_settings={},
            representation_settings={},
            loss_settings={},
            learning_settings={},
            trainer_callbacks=[],
            max_steps_per_iteration=100,
            max_epochs_per_iteration=10,
            representation_cls=PaiNN,
            base_path='',
            training_device=None,
            prediction_device=None,
            db_name='dataset.db',
            transfer_data=None,
            tensorboard=True,
            seed=None,
            **kwargs
    ):
        
        super().__init__(**kwargs)

        self.tensorboard = tensorboard
        if seed is not None:
            seed_everything(seed, workers=True)

        if training_device is None:
            self.training_device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.training_device = torch.device(training_device)

        if prediction_device is None:
            self.prediction_device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        else:
            self.prediction_device = torch.device(prediction_device)

        self.base_path = Path(base_path)

        VERSION_NUMBER = 0
        while True:
            self.version = Path(f'version_{VERSION_NUMBER}_seed_{seed}')
            self.train_path = self.base_path / self.version
            if not self.train_path.is_dir():
                self.train_path.mkdir(parents=True, exist_ok=True)
                break
            else:
                VERSION_NUMBER += 1

        self.db_name = Path(db_name)
        self.data_path = self.train_path / self.db_name

        self.trainer_iteration = 1
            
        self.cutoff = cutoff
        self.dataset_settings = {**dataset_settings,
                                 **{
                                     'batch_size': 16,
                                     'num_train': 0.8,
                                     'num_val': 0.2,
                                     'transforms': [
                                         trn.ASENeighborList(cutoff=cutoff),
                                         trn.CastTo32()
                                     ],    
                                     'num_workers': 8,
                                     'pin_memory': True,
                                     'split_file': None #str(self.train_path / Path('split.npz')),
                                 }
        }
        self.representation_settings = {**representation_settings,
                                        **{
                                            'n_atom_basis': 96,
                                            'n_interactions': 5,
                                            'radial_basis': spk.nn.GaussianRBF(n_rbf=20, cutoff=cutoff),
                                            'cutoff_fn': spk.nn.CosineCutoff(cutoff)
                                        }
        }
        self.loss_settings = {**loss_settings,
                              **{
                                  'energy_weight': 0.01,
                                  'forces_weight': 0.99
                              }
        }
        self.learning_settings = {**learning_settings,
                                  **{
                                      'optimizer_cls': torch.optim.AdamW,
                                      'optimizer_args': {"lr": 1e-4},
                                      'scheduler_cls': spk.train.ReduceLROnPlateau,
                                      'scheduler_args': {'factor': 0.5, 'patience': 1000, 'verbose': True},
                                      'scheduler_monitor': 'val_loss',
                                  }
        }

        self.ckpt_path = self.train_path / 'last.ckpt'
        self.callbacks = trainer_callbacks + [pl.callbacks.ModelCheckpoint(dirpath=str(self.train_path), filename='last', save_last=True)]
        # self.callbacks = [
        #     spk.train.ModelCheckpoint(
        #         model_path=str(self.train_path / Path('best_inference_model')),
        #         monitor='val_loss',
        #         save_top_k=-1,
        #         save_last=True,
        #         every_n_epochs=1,
        #     ),
        #     pl.callbacks.LearningRateMonitor(logging_interval='epoch')
        # ]
        
        self.max_steps_per_iteration = max_steps_per_iteration
        self.max_epochs_per_iteration = max_epochs_per_iteration
        self.representation_cls = representation_cls

        
        # Training DB
        if self.data_path.is_file():
            self.writer('ASE Database already exist. \n Connecting to existing database.')
            self.spk_database = ASEAtomsData(datapath=str(self.data_path))
        else:
            self.spk_database = ASEAtomsData.create(
                str(self.data_path),
                distance_unit = 'Ang',
                property_unit_dict = {
                    'energy': 'eV',
                    'forces': 'eV/Ang'
                }
            )

        # Transfer data
        self.transfer_data = transfer_data
        if self.transfer_data is not None:        
            self.add_data(self.transfer_data)
        

        # Model
        representation = self.representation_cls(**self.representation_settings)
        pred_energy = spk.atomistic.Atomwise(n_in=self.representation_settings.get('n_atom_basis'),
                                             output_key='energy')
        pred_forces = spk.atomistic.Forces(energy_key='energy', force_key='forces')

        pairwise_distance = spk.atomistic.PairwiseDistances()
        
        self.nnpot = spk.model.NeuralNetworkPotential(
            representation=representation,
            input_modules=[pairwise_distance],
            output_modules=[pred_energy, pred_forces],
            postprocessors=[
                trn.CastTo64(),
            ]
        )

        # Output
        output_energy = spk.task.ModelOutput(
            name='energy',
            loss_fn=torch.nn.MSELoss(),
            loss_weight=self.loss_settings.get('energy_weight'),
            metrics={
                "MAE": torchmetrics.MeanAbsoluteError()
            }
        )

        output_forces = spk.task.ModelOutput(
            name='forces',
            loss_fn=torch.nn.MSELoss(),
            loss_weight=self.loss_settings.get('forces_weight'),
            metrics={
                "MAE": torchmetrics.MeanAbsoluteError()
            }
        )

        self.task = spk.task.AtomisticTask(
            model=self.nnpot,
            outputs=[output_energy, output_forces],
            **self.learning_settings,
        )

        # Logging
        if self.tensorboard:
            self.logger = pl.loggers.TensorBoardLogger(save_dir=str(self.base_path),
                                                       name=None, version=str(self.version))
        else:
            self.logger = False


        ########## FOR PREDICTION ##########
        self.converter = AtomsConverter(
            neighbor_list=trn.ASENeighborList(cutoff=self.cutoff),
            device=self.prediction_device.type,
            dtype=torch.float32,
            
        )
            
        self.energy_key = 'energy'
        self.forces_key = 'forces'

        
            

    @property
    def transfer_data(self):
        return self._transfer_data
        
    @transfer_data.setter
    def transfer_data(self, l):
        if isinstance(l, list):
            self._transfer_data = l
            self._transfer_weights = np.ones(len(l))
        elif isinstance(l, dict):
            self._transfer_data = []
            self._transfer_weights = np.array([])
            for key, val in l.items():
                self._transfer_data += val
                self._transfer_weights = np.hstack((self._transfer_weights, float(key) * np.ones(len(val)) ))
        else:
            self._transfer_data = []
            self._trasfer_weights = np.array([])

    @property
    def transfer_weights(self):
        return self._transfer_weights

    
    def set_verbosity(self, verbose):
        self.verbose = verbose
    
    
    def calculate(self, atoms=None, properties=['energy', 'forces'], system_changes=all_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)#, dtype=torch.float64)

        # Something inside schnetpack is checking for atoms objects so converting here. 
        atoms = Atoms(numbers=atoms.get_atomic_numbers(), positions=atoms.positions, cell=atoms.cell, pbc=atoms.pbc)
        
        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)
        if 'energy' in properties:
            e = model_results['energy'].cpu().data.numpy()[0].astype(np.float64)
            # print('Energy pred:', type(e), e, flush=True)
            self.results['energy'] = e
        
        if 'forces' in properties:
            f = model_results['forces'].cpu().data.numpy().astype(np.float64)
            # print('Forces pred:', type(f), f.shape, flush=True)
            self.results['forces'] = f


    ####################################################################################################################
    # Prediction
    ####################################################################################################################

    def predict_energy(self, atoms=None, X=None, return_uncertainty=False):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)
        
        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)
        
        return model_results['energy'].cpu().data.numpy()[0].astype(np.float64)
        

    def predict_energies(self, atoms_list):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)        
        
        model_inputs = self.converter(atoms_list)
        model_results = self.nnpot(model_inputs)

        return model_results['energy'].cpu().data.numpy().astype(np.float64)
            

    def predict_uncertainty(self, atoms=None, X=None, k=None):
        self.writer('Uncertainty not implemented.')
        return 0.

        
    def predict_local_energy(self, atoms=None, X=None):
        self.writer('Local energy not implemented.')
        return np.zeros((len(atoms), ))


    def predict_forces(self, atoms, return_uncertainty=False, **kwargs):
        self.nnpot.eval()
        self.nnpot.to(device=self.prediction_device.type)
        
        model_inputs = self.converter(atoms)
        model_results = self.nnpot(model_inputs)

        return model_results['forces'].cpu().data.numpy()


    @agox_writer
    def train_model(self, training_data, **kwargs):
        self.writer('Training PaiNN model')
        self.nnpot.train()
        self.nnpot.to(device=self.training_device.type)

        self.ready_state = True
        self.atoms = None

        self.add_data(training_data)

        # Dataloader
        dataset = AtomsDataModule(self.data_path, **self.dataset_settings)
        dataset.prepare_data()
        dataset.setup()

        
        trainer = pl.Trainer(
            accelerator=self.training_device.type,
            # devices=1,
            callbacks=self.callbacks,
            logger=self.logger,
            default_root_dir=str(self.train_path),
            max_epochs=self.trainer_iteration*self.max_epochs_per_iteration,
            max_steps = self.trainer_iteration*self.max_steps_per_iteration,
            enable_progress_bar=False,
            enable_model_summary=False,
            
        )
        ckpt_path = str(self.ckpt_path) if self.trainer_iteration > 1 else None
        trainer.fit(self.task, datamodule=dataset, ckpt_path=ckpt_path)
        
        self.trainer_iteration += 1



            
    @agox_writer
    def update_model(self, new_data, all_data):
        self.add_data(new_data)
        self.train_model([])
                
                 
    ####################################################################################################################
    # Assignments:
    ####################################################################################################################

    @agox_writer
    @Observer.observer_method        
    def training_observer(self, database, state):
        iteration = state.get_iteration_counter()

        if iteration < self.iteration_start_training:
            return
        if (iteration % self.update_period != 0) * (iteration != self.iteration_start_training):
            return

        all_data = database.get_all_candidates()
        self.writer(f'length all data: {len(all_data)}')
        
        if self.ready_state:
            full_update = False
            data_amount_before = len(self.spk_database) - len(self.transfer_data)
            data_for_training = all_data
            data_amount_new = len(data_for_training) - data_amount_before
            new_data = data_for_training[-data_amount_new:] 
        else:
            full_update = True
            data_for_training = all_data

        if full_update:
            self.train_model(data_for_training)
        else:
            self.update_model(new_data, data_for_training)
        

    
    def load(self, path):
        state_dict = torch.load(path, map_location=self.prediction_device.type).state_dict()
        self.nnpot.load_state_dict(state_dict)
        


    def add_data(self, data_list):
        if len(data_list) == 0:
            return
        
        property_list = []
        for a in data_list:
            e = a.get_potential_energy(apply_constraint=False)
            f = a.get_forces(apply_constraint=False).reshape(-1, 3)
            c = SPC(a, energy=e, forces=f)
            a.set_calculator(c)
            properties = {'energy': np.array([e]), 'forces': f}
            property_list.append(properties)
        self.spk_database.add_systems(property_list, data_list)





import nbformat as nbf
import pathlib
import subprocess

def add_notebook_parser(subparsers):
    parser = subparsers.add_parser('automagic', help='Create a Jupyter notebook for search analysis.')
    parser.set_defaults(func=cli_notebook)

    parser.add_argument('directories', nargs='+', type=str) # List - Directories
    parser.add_argument('-n', '--name', type=str, default='automagic_analysis') # Name of notebook file (without .ipynb)
    parser.add_argument('-ao', '--auto_open', action='store_false') # Autopen notebook in VS code

def cli_notebook(args):
    directories = args.directories
    auto_open = args.auto_open
    name = args.name

    # Need to massage the input of directories a little bit:
    for i in range(len(directories)):
        directories[i] = "'" + directories[i] + "'"

    # Convenience settings:
    figsize = 'figsize = (7, 7)'

    # Create notebook object
    nb = nbf.v4.new_notebook()

    # Add functions:
    markdown = nbf.v4.new_markdown_cell
    code = nbf.v4.new_code_cell

    # List to hold cell contents
    cells = nb['cells'] = []

    def add_cell(content_string, add_function):
        cells.append(add_function(content_string))

    # Create initial text
    this_file = pathlib.Path(__file__).absolute()
    headline = """
# Automagic AGOX Analysis
This is an automatically generated AGOX analysis notebook.\n 
Notebook generation is defined in this file: {}
    """.format(this_file)
    #cells.append(markdown(headline))
    add_cell(headline, markdown)


    # Import statements % basic settings
    import_block = """import numpy as np
import ipywidgets as widgets
import matplotlib.pyplot as plt
from agox.utils.batch_analysis import Analysis
from agox.utils.jupyter_interactive import InteractiveStructureHistogram, InteractiveSuccessStats, InteractiveEnergy, sorted_species"""
    #cells.append(code(import_block))
    add_cell(import_block, code)

    # Load block: 
    load_headline = """## Scan and read db-files in the directories and calculate CDF.
This is the block that takes the most times, avoid rerunning unless changes have been made that require re-reading the 
directories, such as adding additional ones. """
    add_cell(load_headline, markdown)

    load_block = """
analysis = Analysis()
force_reload = False
"""
    for directory in directories:
        load_block += 'analysis.add_directory({}, force_reload=force_reload)\n'.format(directory)
    load_block += 'analysis.compile_information()\n'
    load_block += 'analysis.calculate_CDF()'

    #cells.append(code(load_block))
    add_cell(load_block, code)

    # Success statistics block:
    succes_headline = """
## Analysis
    """
    add_cell(succes_headline, markdown)


    success_block = """ISS = InteractiveSuccessStats(analysis)
out = widgets.interactive_output(ISS.update_plot, ISS.widget_dict)
widgets.HBox([widgets.VBox(ISS.widget_list), out])""".format(figsize)
    add_cell(success_block, code)

    energy_int_block = """IE = InteractiveEnergy(analysis)
out = widgets.interactive_output(IE.update_plot, IE.widget_dict)
widgets.HBox([widgets.VBox(IE.widget_list), out])"""
    add_cell(energy_int_block, code)


    interactive_struct_hist = """ISH = InteractiveStructureHistogram(analysis)
num_structures = np.sum(analysis.restarts)
index = widgets.IntSlider(min=0, max=num_structures-1, value=0, description='Index')
widgets.interactive(ISH.update_plot, index=index)"""
    add_cell(interactive_struct_hist, code)


    structure_viewer = """from ase.visualize import view
structures, energies = analysis.get_best_structures()
structures = [sorted_species(atoms) for atoms in structures]
view(structures, viewer='ngl')"""
    add_cell(structure_viewer, code)

    # Write notebook:
    notebook_path = name + '.ipynb'
    nbf.write(nb, notebook_path)

    # This will only work with appropriately configured VS-code installation.
    if auto_open:
        subprocess.run('code {}'.format(notebook_path), shell=True)






import numpy as np
from agox.environments import Environment
from agox.generators import RandomGenerator
from ase import Atoms

template = Atoms("", cell=np.eye(3) * 12)
confinement_cell = np.eye(3) * 6
confinement_corner = np.array([3, 3, 3])
environment = Environment(
    template=template,
    symbols="Au8Ni8",
    confinement_cell=confinement_cell,
    confinement_corner=confinement_corner,
)

random_generator = RandomGenerator(
    **environment.get_confinement())

candidate = random_generator(sampler=None, 
                            environment=environment)

print(f'{candidate = }')


from agox.models.descriptors.fingerprint_jax.utils import (
    AtomsData,
    AtomsConverter,
    get_pair_index,
    get_triplet_index,
    compute_radial_features,
    compute_angular_features,
    compute_radial_feature_grad,
    compute_angular_feature_grad,
)

import jax
import jax.numpy as jnp
import numpy as np

from itertools import product, combinations_with_replacement
from ase import Atoms

from agox.models.descriptors import DescriptorBaseClass
from typing import Dict

from agox.candidates import StandardCandidate


class JaxFingerprint(DescriptorBaseClass):
    """
    Fingerprint descriptor

    Parameters:
    -----------
    r_cut_radial: float
        Cutoff for radial features.
    r_cut_angular: float
        Cutoff for angular features.
    n_bins_radial: int
        Number of bins for radial features.
    n_bins_angular: int
        Number of bins for angular features.
    sigma_radial: float
        Gaussian width for radial features.
    sigma_angular: float
        Gaussian width for angular features.
    gamma: int
        Parameter in angular cutoff function.
    eta: float
        Ratio of angular/radial features.
    n_max_neighbors: int
        Maxinum number of neighbors (to keep a constant shape of neighbor_list)
            Use len(atoms) for molecules
            Default 30 should work for periodic systems
            Increase this value when there is a zero padding error
            Set to -1 to turn this off
    use_angular: bool
        Whether angular features are calculated.
    """

    descriptor_type = "global"
    name = "JaxFingerprint"

    def __init__(
        self,
        r_cut_radial: float = 6.0,
        r_cut_angular: float = 6.0,
        n_bins_radial: int = 30,
        n_bins_angular: int = 30,
        sigma_radial: float = 0.2,
        sigma_angular: float = 0.2,
        gamma: int = 2,
        eta: float = 20,
        pad=True,
        n_max_neighbors: int = 30,
        use_angular: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.r_cut_radial = r_cut_radial
        self.n_bins_radial = n_bins_radial
        self.sigma_radial = sigma_radial

        self.r_cut_angular = r_cut_angular
        self.n_bins_angular = n_bins_angular
        self.sigma_angular = sigma_angular

        self.gamma = gamma
        self.eta = eta
        self.n_max_neighbors = n_max_neighbors
        self.use_angular = use_angular

        self.max_pair = 0
        self.max_triplet = 0

        self.initialize_features(self.environment.get_template())

        self.atoms_converter = AtomsConverter(r_cut=r_cut_radial, collect_triplet=use_angular, 
                                              r_cut_angular=r_cut_angular, pad=pad)

    def get_number_of_centers(self, atoms: Atoms) -> int:
        return 1

    def initialize_features(self, template: Atoms) -> None:
        """Compute unique pairs, triples and feature dimensions
        (called when the stoichiometry changes, no jax here)

        Parameters:
            template: an ASE Atoms object contains information about atomic_numbers, pbc, ...
        """

        periodic = template.get_pbc().any()
        numbers = template.get_atomic_numbers()
        unique, counts = np.unique(numbers, return_counts=True)

        # pairs, the center atom i does not matter
        if periodic:
            pairs = np.array(list(combinations_with_replacement(unique, 2)))
        else:
            pairs = []
            for z_ij in combinations_with_replacement(unique, 2):
                n_valid_number = 0
                for z, c in zip(unique, counts):
                    if z_ij.count(z) <= c:
                        n_valid_number += 1
                if n_valid_number == len(unique):
                    pairs.append(z_ij)
            pairs = np.array(pairs)
        self.pairs = np.sum(pairs * [10, 1], axis=1)

        # triplets, the center atom i matters
        if self.use_angular:
            if periodic:
                triplets = np.array(list(product(unique, repeat=3)))
                triplets = triplets[
                    triplets[:, 1] <= triplets[:, 2]
                ]  # i,j,k is identical to i,k,j
            else:
                triplets = []
                for z_ijk in product(unique, repeat=3):
                    if z_ijk[1] <= z_ijk[2]:
                        n_valid_number = 0
                        for z, c in zip(unique, counts):
                            if z_ijk.count(z) <= c:
                                n_valid_number += 1
                        if n_valid_number == len(unique):
                            triplets.append(z_ijk)
                triplets = np.array(triplets)
                print(triplets)
            self.triplets = np.sum(triplets * [100, 10, 1], axis=1)

    def create_features(self, atoms: Atoms, inputs: Dict = None) -> jax.Array:
        if inputs is None:
            inputs = self.converter(atoms)

        features = self._create_radial_features(inputs)

        if self.use_angular:
            angular_features = self._create_angular_features(inputs)
            features = jnp.concatenate((features, angular_features))

        # Reshape to [n_centers, n_features]
        features = features.reshape((self.get_number_of_centers(atoms), -1))

        return features

    def create_feature_gradient(self, atoms: Atoms, inputs: Dict = None) -> jax.Array:
        if inputs is None:
            inputs = self.converter(atoms)

        features_grad = self._create_radial_feature_gradients(inputs)

        if self.use_angular:
            angular_features_grad = self._create_angular_feature_gradients(inputs)
            features_grad = jnp.concatenate((features_grad, angular_features_grad))

        # Reshape to [n_centers, n_atoms, 3, n_features]
        features_grad = features_grad.reshape(
            (self.get_number_of_centers(atoms), len(atoms), 3, -1)
        )

        return features_grad

    @StandardCandidate.cache('inputs')
    def converter(self, atoms: Atoms) -> Dict:
        inputs = self.atoms_converter.convert(atoms)

        prefactor_radial = atoms.get_volume() / (
            4 * np.pi * len(atoms) ** 2 * self.r_cut_radial / self.n_bins_radial
        )
        inputs.prefactor_radial = prefactor_radial
        prefactor_angular = self.eta * atoms.get_volume() / len(atoms) ** 3
        inputs.prefactor_angular = prefactor_angular

        return inputs

    @classmethod
    def from_atoms(cls, atoms, **kwargs):
        """
        Create a descriptor from an atoms object.

        Parameters
        ----------
        atoms : Atoms object
            The atoms object used to setup the descriptor.
        kwargs : dict
            Any keyword arguments to pass to the descriptor.
        """
        from agox.environments import Environment

        environment = Environment(
            template=atoms,
            symbols="",
            use_box_constraint=False,
            print_report=False,
            confinement_cell=atoms.get_cell(),
        )
        return cls(environment=environment, **kwargs)

    def _create_radial_features(self, inputs: AtomsData) -> jax.Array:
        idx_pair = get_pair_index(
            self.pairs, inputs.Z[inputs.idx_i], inputs.Z[inputs.idx_j]
        )

        radical_features = compute_radial_features(
            inputs.R,
            inputs.idx_i,
            inputs.idx_j,
            idx_pair,
            inputs.cell_shift,
            inputs.cell,
            len(self.pairs),
            self.r_cut_radial,
            self.n_bins_radial,
            self.sigma_radial,
            inputs.prefactor_radial,
        )

        return radical_features

    def _create_radial_feature_gradients(self, inputs: AtomsData) -> jax.Array:
        idx_pair = get_pair_index(
            self.pairs, inputs.Z[inputs.idx_i], inputs.Z[inputs.idx_j]
        )

        radical_features_grad = compute_radial_feature_grad(
            inputs.R,
            inputs.idx_i,
            inputs.idx_j,
            idx_pair,
            inputs.cell_shift,
            inputs.cell,
            len(self.pairs),
            self.r_cut_radial,
            self.n_bins_radial,
            self.sigma_radial,
            inputs.prefactor_radial,
        )

        return radical_features_grad

    def _create_angular_features(self, inputs: AtomsData) -> jax.Array:
        idx_triplet_i = inputs.idx_i[inputs.idx_triplet_ij]
        idx_triplet_j = inputs.idx_j[inputs.idx_triplet_ij]
        idx_triplet_k = inputs.idx_j[inputs.idx_triplet_ik]

        idx_triplet = get_triplet_index(
            self.triplets,
            inputs.Z[idx_triplet_i],
            inputs.Z[idx_triplet_j],
            inputs.Z[idx_triplet_k],
        )

        angular_features = compute_angular_features(
            inputs.R,
            inputs.idx_i,
            inputs.idx_j,
            inputs.idx_triplet_ij,
            inputs.idx_triplet_ik,
            idx_triplet,
            inputs.cell_shift,
            inputs.cell,
            len(self.triplets),
            self.r_cut_angular,
            self.n_bins_angular,
            self.sigma_angular,
            self.gamma,
            inputs.prefactor_angular,
        )

        return angular_features

    def _create_angular_feature_gradients(self, inputs: AtomsData) -> jax.Array:
        idx_triplet_i = inputs.idx_i[inputs.idx_triplet_ij]
        idx_triplet_j = inputs.idx_j[inputs.idx_triplet_ij]
        idx_triplet_k = inputs.idx_j[inputs.idx_triplet_ik]

        idx_triplet = get_triplet_index(
            self.triplets,
            inputs.Z[idx_triplet_i],
            inputs.Z[idx_triplet_j],
            inputs.Z[idx_triplet_k],
        )

        angular_features_grad = compute_angular_feature_grad(
            inputs.R,
            inputs.idx_i,
            inputs.idx_j,
            inputs.idx_triplet_ij,
            inputs.idx_triplet_ik,
            idx_triplet,
            inputs.cell_shift,
            inputs.cell,
            len(self.triplets),
            self.r_cut_angular,
            self.n_bins_angular,
            self.sigma_angular,
            self.gamma,
            inputs.prefactor_angular,
        )

        return angular_features_grad


if __name__ == "__main__":
    # from ase.build import molecule

    from ase import Atoms
    from ase.io import read
    from tqdm import tqdm

    all_atoms = read('/home/machri/PythonPackages/agox-v2/agox/test/datasets/B12-dataset.traj', ':')
    use_angular = True
    eta = 5
    r_cut_all = 5.0
    n_bins = 30
    descriptor = JaxFingerprint.from_atoms(
        all_atoms[0],
        r_cut_radial=r_cut_all,
        r_cut_angular=r_cut_all,
        n_bins_radial=n_bins,
        use_angular=use_angular,
        eta=eta,
        n_max_neighbors=100,
    )

    # for atoms in tqdm(all_atoms):
    #     features = descriptor.get_features(atoms)
    #     # gradients = descriptor.get_feature_gradient(atoms)

    from agox.models.GPR import GPR
    from agox.models.GPR.kernels import RBF, Constant as C, Noise

    kernel = C(1) * RBF(1) + Noise()
    model = GPR(descriptor=descriptor, kernel=kernel, use_ray=False, n_optimize=0)

    model.train(all_atoms[:5])

    E = model.predict_energy(all_atoms[5])
    print(E)


from agox.evaluators.local_optimization import LocalOptimizationEvaluator

class SinglePointEvaluator(LocalOptimizationEvaluator):

    """
    Does a single point calculation of a candidate.

    Bases on the LocalOptimizationEvaluator, but with steps=0.

    Parameters:
    -----------
    calculator: ASE calculator
        The calculator to use for the evaluation.
    """

    name = 'SinglePointEvaluator'

    def __init__(self, calculator, **kwargs): 
        super().__init__(calculator, optimizer_run_kwargs=dict(steps=0), **kwargs)


from agox.test.test_utils import TemporaryFolder, test_folder_path, check_folder_is_empty

def test_tracker(tmp_path, cmd_options):

        with TemporaryFolder(tmp_path):
            from agox.tracker import Tracker
            import os
            import matplotlib
            matplotlib.use('Agg')
            import numpy as np
            from agox import AGOX
            from agox.databases import Database
            from agox.environments import Environment
            from agox.evaluators import LocalOptimizationEvaluator
            from agox.generators import RandomGenerator
            from ase import Atoms
            from argparse import ArgumentParser

            # Manually set seed and database-index
            seed = 42
            database_index = 0

            # Using argparse if e.g. using array-jobs on Slurm to do several independent searches. 
            # parser = ArgumentParser()
            # parser.add_argument('-i', '--run_idx', type=int, default=0)
            # args = parser.parse_args()

            # seed = args.run_idx
            # database_index = args.run_idx

            ##############################################################################
            # Calculator
            ##############################################################################

            from ase.calculators.emt import EMT

            calc = EMT()

            ##############################################################################    
            # System & general settings:
            ##############################################################################
                
            template = Atoms('', cell=np.eye(3)*12)
            confinement_cell = np.eye(3) * 6
            confinement_corner = np.array([3, 3, 3])
            environment = Environment(template=template, symbols='Au8Ni8', 
                confinement_cell=confinement_cell, confinement_corner=confinement_corner)

            # Database
            db_path = 'db{}.db'.format(database_index) # From input argument!
            database = Database(filename=db_path, order=3)

            ##############################################################################
            # Search Settings:
            ##############################################################################
                
            random_generator = RandomGenerator(**environment.get_confinement(), 
                environment=environment, order=1)

            random_generator.add_tracked_attribute('set_key')
            key = f'{random_generator.__name__}.set_key'

            # Wont relax fully with steps:5 - more realistic setting would be 100+.
            evaluator = LocalOptimizationEvaluator(calc, gets={'get_key':'candidates'}, 
                optimizer_run_kwargs={'fmax':0.05, 'steps':5}, store_trajectory=False,
                order=2, constraints=environment.get_constraints())

            ##############################################################################
            # Let get the show running! 
            ##############################################################################

            agox = AGOX(random_generator, database, evaluator, seed=seed)

            # State tracking: 
            def mean_position(list_of_atoms):
                positions = np.array([atoms.get_positions() for atoms in list_of_atoms])
                return positions.mean()

            # Track the mean position of the atoms in the system.
            agox.tracker.add_state_tracker(mean_position, 'mean_position', 'evaluated_candidates')
            agox.tracker.add_state_tracker(mean_position, 'mean_position', 'candidates')

            agox.run(N_iterations=10)

            loaded_properties = Tracker.load('db0_tracker.npz')

        tracker = agox.tracker

        assert key in tracker.tracked_attributes.keys()

        for key in tracker.tracked_attributes.keys():
            assert key in loaded_properties.keys()
            assert (tracker.tracked_attributes[key] == loaded_properties[key]).all()
            assert len(tracker.tracked_attributes[key]) == 10 # 10 iterations



import pytest
from agox.test.test_utils import environment_and_dataset
from agox.postprocessors import CenteringPostProcess, WrapperPostprocess, RelaxPostprocess, ParallelRelaxPostprocess

from agox.models.descriptors.fingerprint import Fingerprint
from agox.models.GPR.kernels import RBF, Constant as C, Noise
from agox.models.GPR import GPR

import numpy as np

def base_setup(environment, dataset):
    return {}

def make_model(environment, dataset):
    model = GPR(descriptor=Fingerprint(environment=environment),
                kernel=C() * RBF() + Noise(0.01),
                database=None)
    model.train(dataset)
    return {'model':model}

@pytest.mark.ray
@pytest.mark.parametrize('postprocess_class, setup_kwargs, setup_func', [
    (CenteringPostProcess, {}, base_setup),
    (WrapperPostprocess, {}, base_setup),
    (RelaxPostprocess, {'optimizer_run_kwargs':{'steps':10}}, make_model),
    (ParallelRelaxPostprocess, {'optimizer_run_kwargs':{'steps':10}}, make_model)

])
def test_postprocess(postprocess_class, setup_kwargs, setup_func, environment_and_dataset):
    environment, dataset = environment_and_dataset
    postprocessor = postprocess_class(**setup_kwargs, **setup_func(environment, dataset))

    try:
        postprocessor.postprocess(dataset[0])
    except NotImplementedError:
        postprocessor.process_list(dataset[0:8])





import numpy as np 
from time import time
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
import sqlite3
import os
import sys
import os.path
if sys.version >= '3':
    buffer = memoryview

T2000 = 946681200.0  # January 1. 2000
YEAR = 31557600.0  # 365.25 days


def now():
    """Return time since January 1. 2000 in years."""
    return (time() - T2000) / YEAR

def nothing(a):
    return a

def blob(a):
    return buffer(np.ascontiguousarray(a))


def deblob(buf, dtype=np.float64, shape=None):
    """Convert blob/buffer object to ndarray of correct dtype and shape.
    (without creating an extra view)."""

    if buf is None:
        return None
    if len(buf) == 0:
        array = np.zeros(0, dtype)
    else:
        if len(buf) % 2 == 1 and False:
            # old psycopg2:                                                                                     
            array = np.fromstring(str(buf)[1:].decode('hex'), dtype)
        else:
            array = np.frombuffer(buf, dtype)
        if not np.little_endian:
            array.byteswap(True)
    if shape is not None:
        array.shape = shape
    return array


def export_candidates(dbs, name='run', save=False):
    ebests = []
    emin = float(1E10)
    es = []
    best_structure = None
    for i, d in enumerate(dbs):
        e = []
        ebest = []
        cands = d.get_all_structures_data()
        for c in cands:
            # Extract candidate energy:
            E = c['energy'] if c['energy'] is not None else np.nan
            e.append(E)
            if len(ebest) == 0:
                ebest.append(E)
            else:
                ebest.append(np.nanmin([ebest[-1], e[-1]]))
            if e[-1] < emin:
                emin = e[-1]
                cell = c['cell']
                num = c['type']
                pos = c['positions']
                pbc = c['pbc']
                best_structure = Atoms(symbols = num,
                                       positions = pos,
                                       cell = cell,
                                       pbc = pbc)
                calc = SinglePointCalculator(best_structure, energy = e[-1])
                best_structure.set_calculator(calc)                
        ebests.append(ebest)
        es.append(e)
        if save:
            np.save(name + str(i) + '_Energies.npy',(e, ebest))

    return es, ebests, best_structure


def db_to_ase(cand):
    """

    Converts a database representation (dictionary) of a structur to an ASE atoms object

    Parameters
    ----------
    cand :  database representation of a structure

    Returns
    -------
    struc : ASE Atoms object

    """

    e = cand['energy']
    try:
        f = cand['forces']
    except:
        f = 0
    #f = cand.get('forces',None)
    pos = cand['positions']
    num = cand['type']
    cell = cand['cell']
    struc = Atoms(symbols = num,
                  positions = pos,
                  cell = cell)
    calc = SinglePointCalculator(struc, energy=e, forces=f)
    struc.set_calculator(calc)

    return struc



from typing import List
from agox.utils import candidate_list_comprehension
from ase import Atoms
import numpy as np
from agox.models.GPR.sGPR import SparseGPR

class SparseGPREnsemble(SparseGPR):

    """
    Sparse Gaussian Process Regression with ensemble of models.

    Parameters
    ----------
    N_ensemble : int
        Number of models in the ensemble.
    target_noise : float
        Noise added to the target values pr. atom
    prior_noise : float
        Noise added to the prior values pr. atom    
    """

    name = 'SparseGPREnsemble'

    def __init__(self, N_ensemble=5, target_noise=0.25, prior_noise=0, *args, **kwargs):
        self.priors_per_atom = None
        self.ensemble_alpha = None
        self.N_ensemble = N_ensemble
        self.target_noise = target_noise
        self.prior_noise = prior_noise

        super().__init__(*args, **kwargs)
        for attr in ["priors_per_atom", "ensemble_alpha"]:
            self.add_dynamic_attribute(attr)

    def _preprocess(self, data: List[Atoms]):
        X, y = super()._preprocess(data)

        all_data = self.get_all_data(data)

        # Repeat y in the ensemble dimension
        y = np.repeat(y, self.N_ensemble + 1, axis=1)

        # Add noise to y in the ensemble dimension:
        centers = np.array([self.descriptor.get_number_of_centers(atoms) for atoms in all_data]).reshape(-1, 1)
        target_noise = np.repeat(self.target_noise * centers, self.N_ensemble, axis=1)
        random_noise = np.random.normal(scale=target_noise, size=y[:self._N_energies, 1:].shape)
        y[:self._N_energies, 1:] += random_noise

        # Calculate prior for each model:
        self.priors_per_atom = np.random.normal(scale=self.prior_noise, size=(1, y.shape[1]))
        self.priors_per_atom[0, 0] = 0    
        y -= centers * self.priors_per_atom
        return X, y
    
    def _train_model(self, data: List[Atoms], synchronize=True) -> None:
        super()._train_model(data, synchronize=False, train_projected_process=False)
        self.ensemble_alpha = self.alpha[:, 1:].copy()
        self.alpha = self.alpha[:, 0]

        if self.use_ray and synchronize:
            self.pool_synchronize(
                attributes=["X", "alpha", "ensemble_alpha", "kernel", "priors_per_atom", "mean_energy"], writer=self.writer)

        
    @candidate_list_comprehension
    def predict_uncertainty(
        self, atoms: Atoms, k: np.ndarray = None, x: np.ndarray = None, **kwargs
    ) -> float:

        if self.ensemble_alpha is not None:

            k = self._get_kernel_vector(k, x, atoms)
            centers = self.descriptor.get_number_of_centers(atoms)
            e_pred = np.sum(k.T @ self.ensemble_alpha, axis=0) + centers * self.priors_per_atom[0, 1:]
            return np.std(e_pred)
        else:
            return 0.0

    def predict_variances(self, atoms: Atoms, x: np.ndarray = None, k: np.ndarray = None, **kwargs) -> float:
        if self.ensemble_alpha is not None:
            k = self._get_kernel_vector(k, x, atoms)
            centers = self.descriptor.get_number_of_centers(atoms)
            e_pred = k.T @ self.ensemble_alpha + centers * self.priors_per_atom[0, 1:]
            return e_pred.var(axis=1)
        else:
            return 0.0
    
    @candidate_list_comprehension
    def predict_uncertainty_forces(self, atoms, x=None, k=None, dk_dr=None, **kwargs):

        if k is None:
            if x is None:
                x = self._get_features(atoms)
            k = self.kernel(self.Xm, x)

        if dk_dr is None:
            dk_dr = self._get_kernel_derivative(atoms, x=x)

        # Predict the forces:
        forces = (-dk_dr.sum(axis=0) @ self.ensemble_alpha).squeeze()

        centers = self.descriptor.get_number_of_centers(atoms)
        energies = np.sum(k.T @ self.ensemble_alpha, axis=0) + centers * self.priors_per_atom[0, 1:]
        std = np.std(energies)

        std_forces = (1/(self.N_ensemble*std) * (energies - energies.mean()) * (forces - forces.mean())).sum(axis=2)

        return std_forces









from agox.generators.ABC_generator import GeneratorBaseClass
import numpy as np
from scipy.spatial.distance import cdist
from ase.neighborlist import NeighborList
from scipy.optimize import minimize
from ase import Atoms
from ase.constraints import FixAtoms
from ase.optimize import BFGS

class ComplementaryEnergyGenerator(GeneratorBaseClass):

    """
    Use complementary energy landscape to generate new candidates.

    Parameters
    ------------
    calc: CE calculator
        Used to evaluate the CE energy and forces
    descriptor: Local descriptor
        Used to calculate local descriptor for atoms in structures
    attractor_method: Method for picking attractors
        Used to construct attractors in CE expression
    move_all: Bool
        Whether to move all non-template atoms or not. If "mover_indices"
        is not None, the atoms with the corresponding indices will be free
        to move, thereby ignoring the "move_all" argument.
    mover_indices: List of integers
        Only atoms with the corresponding indices can move during
        candidate generation.
    """
    name = 'ComplementaryEnergyGenerator'

    def __init__(self,
                calculator,
                descriptor,
                attractor_method,
                move_all = True, 
                mover_indices = None,
                replace = True,
                **kwargs):

        super().__init__(replace=replace,**kwargs)
        self.calc = calculator
        self.descriptor = descriptor
        self.attractor_method = attractor_method
        self.move_all = move_all
        self.mover_indices = mover_indices

    def _get_candidates(self, candidate, parents, environment):

        template = candidate.get_template()
        n_template = len(template)

        # Set attractors
        attractors = self.attractor_method.get_attractors(candidate)
        self.calc.attractors = attractors

        # Determine which atoms to move if no indices has been supplied
        # Picks those with highest local complementary energy
        if self.mover_indices is None:
            if self.move_all:
                mover_indices = range(n_template, len(candidate))
            else:
                n_movers = np.random.randint(1, len(candidate) - n_template)
                mover_indices = np.array(range(n_template, len(candidate)))
                self.calc.mover_indices = mover_indices
                local_ce = self.calc.get_local_ce_energy(candidate)
                largest_local_ce_indices = np.argsort(local_ce)[::-1]
                mover_indices = mover_indices[largest_local_ce_indices][:n_movers]

            self.mover_indices = mover_indices

        # Set mover indices and attractors for calculator
        self.calc.mover_indices = self.mover_indices

        # Set calculator
        candidate.calc = self.calc

        def objective_func(pos):
            candidate.positions[self.mover_indices] = np.reshape(pos,[len(self.mover_indices),3])
            ce = candidate.get_potential_energy(apply_constraint = 0)
            grad = - candidate.get_forces(apply_constraint = 0)[self.mover_indices]
            return ce, grad.flatten()

        pos = candidate.positions[self.mover_indices].flatten()
        pos = minimize(objective_func, pos, method = 'BFGS', jac = True, options={'maxiter':75, 'disp':False}).x
        suggested_positions = np.reshape(pos,[len(self.mover_indices), 3])

        for index in range(len(self.mover_indices)):
            i = self.mover_indices[index]
            for _ in range(100):
                if _ == 0:
                    radius = 0
                else:
                    radius = 0.5 * np.random.rand()**(1/self.dimensionality)
                displacement = self.get_displacement_vector(radius)
                suggested_position = suggested_positions[index] + displacement

                # Check confinement limits:
                if not self.check_confinement(suggested_position):
                    continue

                # Check that suggested_position is not too close/far to/from other atoms
                # Skips the atom it self. 
                if self.check_new_position(candidate, suggested_position, candidate[i].number, skipped_indices=[i]):
                    candidate[i].position = suggested_position 
                    break

        return [candidate]

    @classmethod
    def default(cls, environment, database, move_all=True, mover_indices=None, 
                attractors_from_template = False, predefined_attractors = None, **kwargs):

        from agox.models.descriptors.exponential_density import ExponentialDensity
        from agox.generators.complementary_energy.ce_calculators import ComplementaryEnergyDistanceCalculator
        from agox.generators.complementary_energy.attractor_methods.ce_attractors_current_structure import AttractorCurrentStructure

        lambs = [0.5, 1, 1.5]
        rc = 10.
        descriptor = ExponentialDensity(environment=environment, lambs=lambs, rc=rc)
        ce_calc = ComplementaryEnergyDistanceCalculator(descriptor = descriptor)
        ce_attractors = AttractorCurrentStructure(descriptor=descriptor, 
                            attractors_from_template=attractors_from_template, 
                            predefined_attractors=predefined_attractors)
        ce_attractors.attach(database)

        return cls(calculator=ce_calc, descriptor=descriptor, 
                   attractor_method=ce_attractors, move_all=move_all, **environment.get_confinement())

    def get_number_of_parents(self, sampler):
        return 1



import pytest
from agox.test.run_tests.run_utils import agox_test_run

@pytest.mark.ray
def test_pt(tmp_path, cmd_options, ray_fix):
    mode = 'pt' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options)


from agox.test.run_tests.run_utils import agox_test_run
import pytest

@pytest.mark.skip(reason='Not reproducible on CI at the moment.')
def test_gofee_cluster_symmetry(tmp_path, cmd_options, ray_fix):
    mode = 'gofee_cluster_symmetry' # This determines the script that is imported.
    agox_test_run(mode, tmp_path, cmd_options, full_compare=False)



"""
A candidate is the basic piece of data being moved around in the algorithm. 
Generators create candidates, and evaluators evaluate the objective function 
at the coordinates described by the candidate. 

AA
"""

from agox.candidates.ABC_candidate import CandidateBaseClass
from agox.candidates.standard import StandardCandidate

__all__ = ['CandidateBaseClass', 'StandardCandidate']


import re

from agox.generators.ABC_generator import GeneratorBaseClass
from agox.generators.symmetry_utils.symmetry_enforcer import SymmetryEnforcer

from ase.data import covalent_radii
from ase.io import read, write
from ase import Atoms, Atom
from ase.geometry import wrap_positions, get_distances
from ase.build.tools import sort

import numpy as np
import scipy.spatial.distance as ssd
from scipy.cluster.hierarchy import fcluster, linkage
from math import cos, sin, pi, tan, atan

import time


def upper_tri_masking(A):
    m = A.shape[0]
    r = np.arange(m)
    mask = r[:, None] < r
    return A[mask]


class SymmetryGenerator(GeneratorBaseClass):

    """
    Creates random symmetric structures without a seed.

    Parameters:
    -----------
    may_nucleate_at_several_places : bool
        If True, the generator may nucleate at several places in the template, 
        meaning that atoms can independently be put at different places on
        the template. 
    check_fragmented : bool
        If True, the generator will check if the structure is fragmented.
    sym_type : str
        The symmetry type, either 'slab' or 'cluster'.
    force_group : Symmetry group name or None. 
        If not None, the generator will force the symmetry group to be the 
        given group.
    equal_element_list : list
        No idea. 
    not_sym_list : list
        No idea.
    """

    name = "SymmetryGenerator"

    def __init__(
        self,
        may_nucleate_at_several_places=False,
        check_fragmented=False,
        sym_type="slab",
        force_group=None,
        equal_element_list=[],
        not_sym_list=[],
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.may_nucleate_at_several_places = may_nucleate_at_several_places
        self.sym_type = sym_type
        self.check_fragmented = check_fragmented
        # equal element list offers the possibility to switch off element wise deposition for a list of element useful for putting B or N in C60 for instance
        self.equal_element_list = equal_element_list
        self.not_sym_list = not_sym_list
        self.force_group = force_group

    def check_new_position(
        self,
        candidate,
        new_position=None,
        number=None,
        skipped_indices=[],
        mic=True,
        rmax=True,
    ):
        """
        Programing of checking all together
        """
        candi_to_analyse = [
            candidate.positions[i]
            for i in range(len(candidate))
            if i not in skipped_indices
        ]
        numbers = [
            covalent_radii[candidate.numbers[i]]
            for i in range(len(candidate))
            if i not in skipped_indices
        ]
        pos = np.array(candi_to_analyse)
        pos = np.reshape(pos, (-1, 3))
        if new_position is not None:
            new_position = np.array(new_position)
            if new_position.ndim == 2:
                if new_position.shape[0] > 1:
                    number = [number] * new_position.shape[0]
            else:
                new_position = np.reshape(new_position, (-1, 3))
            if isinstance(number, list):
                cov = [covalent_radii[num] for num in number]
                numbers.extend(cov)
            else:
                numbers.append(covalent_radii[number])
            if pos.shape[1] == new_position.shape[1]:
                pos = np.vstack((pos, new_position))
            else:
                return False
        if mic == False:
            pbc = [0, 0, 0]
        else:
            pbc = candidate.pbc
        if np.reshape(pos, (-1, 3)).shape[0] > 1:
            pos = np.reshape(pos, (-1, 3))
            vector, dMat = get_distances(pos, pos, cell=candidate.cell, pbc=pbc)
        else:
            return True
            # return False
            # raise
        dist = upper_tri_masking(dMat)
        dist1 = np.tile(numbers, (len(numbers), 1))
        dist2 = dist1.T
        dist_tot_min = self.c1 * (dist1 + dist2)
        dist_tot_max = self.c2 * (dist1 + dist2)
        vect_dist_min = upper_tri_masking(dist_tot_min)
        vect_dist_max = upper_tri_masking(dist_tot_max)

        # print(vect_dist_max-dist)
        if dist.shape[0] == vect_dist_min.shape[0]:
            new_vec = dist - vect_dist_min
        else:
            return False
        if any(new_vec < 0):
            return False
        if rmax == True:
            if np.any(np.all(dist_tot_max - dMat < 0, axis=0)):
                return False
        return True

    def get_number_of_parents(self, sampler):
        return 0

    def _get_candidates(self, sampler, parents, environment):
        template = environment.get_template()
        self.slab = template
        numbers_list = list(environment.get_numbers())
        len_of_template = len(template)
        candidate = template.copy()
        group = None
        table_mult = []
        self.se = SymmetryEnforcer(
            self.slab,
            group=group,
            mode=self.sym_type,
            confinement_cell=self.confinement_cell,
            cell_corner=self.confinement_corner,
            dimensionality=self.dimensionality,
        )
        tries = 0
        t0 = time.time()
        stop = False

        while len(numbers_list) > 0 and tries < 12 and not stop:
            # one can shuffle the atomic deposition but it is not recommended
            # np.random.shuffle(numbers_list)
            atomic_number = numbers_list[0]
            placing_first_atom = len(candidate) == len_of_template
            for _ in range(30):
                at_in_list = 0
                # the deposition is element wise but it could be switched off with the following part
                for num in numbers_list:
                    if num == atomic_number:
                        at_in_list += 1
                    elif num in [self.equal_element_list] and atomic_number in [
                        self.equal_element_list
                    ]:
                        at_in_list += 1
                self.se.max_at = at_in_list
                if self.force_group is not None:
                    if isinstance(self.force_group, list):
                        self.force_group = np.random.choice(self.force_group)
                    self.se.force_group = self.force_group
                if self.se.group is None:
                    self.se.group = self.se.select_group(mode=self.sym_type)
                suggested_position = None
                if (
                    placing_first_atom or self.may_nucleate_at_several_places
                ):  # If the template is completely empty.
                    suggest = self.get_box_vector()
                    if atomic_number not in self.not_sym_list:
                        if len(candidate) > 0:
                            psym = self.se.get_symmetry_points(
                                suggest,
                                atomic_number=atomic_number,
                                sym_type=self.sym_type,
                                candi=candidate,
                            )
                        else:
                            psym = self.se.get_symmetry_points(
                                suggest,
                                atomic_number=atomic_number,
                                sym_type=self.sym_type,
                            )
                    else:
                        psym = np.array([suggest])
                    if psym is None:
                        tries += 1
                        build_succesful = False
                        if self.se.group["name"].__contains__("C"):
                            stop = True
                            break
                        continue
                    Nadd = psym.shape[0]
                    # > atomic numbers in number_list
                    for p in psym:
                        if suggested_position is None:
                            suggested_position = np.array([p])
                        else:
                            suggested_position = np.vstack((suggested_position, p))
                else:  # Pick only among atoms placed by the generator.
                    placed_atom = candidate[
                        np.random.randint(len_of_template, len(candidate))
                    ]
                    suggest = placed_atom.position.copy()
                    # Get a vector at an appropriate radius from the picked atom.
                    vec = self.get_sphere_vector(atomic_number, placed_atom.number)
                    suggest += vec
                    if atomic_number not in self.not_sym_list:
                        psym = self.se.get_symmetry_points(
                            suggest,
                            atomic_number=atomic_number,
                            sym_type=self.sym_type,
                            candi=candidate,
                        )
                    else:
                        psym = np.array([suggest])
                    if psym is None:
                        tries += 1
                        build_succesful = False
                        if self.se.group["name"].__contains__("C"):
                            break
                        continue
                    Nadd = psym.shape[0]
                    for p in psym:
                        if suggested_position is None:
                            suggested_position = np.array([p])
                        else:
                            suggested_position = np.vstack((suggested_position, p))
                if Nadd > at_in_list or suggested_position is None:
                    tries += 1
                    continue
                check = self.check_confinement(suggested_position)
                if not check.all():
                    tries += 1
                    build_succesful = False
                    continue
                # C2 is not really useful when mapping a slab
                if self.sym_type == "cluster":
                    pos = np.vstack((candidate.positions, suggested_position))
                    vector, distances = get_distances(pos, pos)
                    num = candidate.numbers
                    num = np.concatenate((num, [atomic_number] * psym.shape[0]))
                    cov_dist = np.array([covalent_radii[n] for n in num])
                    relative_dist = distances / np.add.outer(cov_dist, cov_dist)
                    L = self.se.get_laplacian_matrix(relative_dist)
                    w, _ = np.linalg.eig(L)
                if len(candidate) > 0:
                    check_pos = self.check_new_position(
                        candidate, suggested_position, atomic_number, rmax=False
                    )
                if len(candidate) > 0 and self.sym_type == "cluster":
                    check_pos = self.check_new_position(
                        candidate, suggested_position, atomic_number
                    )
                    if (
                        self.check_fragmented
                        and np.sum(np.abs(w) < 1e-12) > 1
                        and (len(candidate) + len(suggested_position))
                        > len(list(environment.get_numbers())) / 2
                        and self.se.system == "cluster"
                    ):
                        tries += 1
                        build_succesful = False
                        continue
                # print(check_pos)
                if len(candidate) == 0 or check_pos:
                    if self.sym_type == "cluster":
                        if (
                            self.check_fragmented
                            and np.sum(np.abs(w) < 1e-12) > 1
                            and len(suggested_position)
                            == len(list(environment.get_numbers()))
                        ):
                            build_succesful = False
                            tries += 1
                            continue
                    build_succesful = True
                    table_mult.append(int(len(suggested_position)))
                    for sugg in suggested_position:
                        numbers_list.remove(atomic_number)
                        candidate.extend(
                            Atoms(numbers=[atomic_number], positions=[sugg])
                        )
                    # tries=0
                    break
                build_succesful = False
        if not build_succesful or len(candidate) != len(template) + len(
            list(environment.get_numbers())
        ):
            # print('Symmetry generator failing at producing valid structure')
            return []
        if self.sym_type == "cluster":
            table_mult = self.se.table_cluster_desym(self.se.group, table_mult)
        # candidate = self.convert_to_candidate_object(candidate, template)
        
        candidate.add_meta_information("description", self.name)
        candidate.add_meta_information("space_group", self.se.group["name"])
        candidate.add_meta_information("table_mult", np.array(table_mult))
        candidate.add_meta_information("non_sym", self.not_sym_list)

        if self.se.group["name"].__contains__("C") is True:
            candidate.add_meta_information("lost_sym", 2)
        if (
            self.se.group["name"].__contains__("D") is True
            and len(self.se.group["name"]) == 2
        ):
            candidate.add_meta_information("lost_sym", 1)
        if self.se.cluster_center is not None:
            candidate.add_meta_information("center", self.se.cluster_center)

        candidate.add_write_key("space_group")
        
        return [candidate]


        # write('sym.vasp', candidate, format='vasp')
        


# Copyright 2021-2022, Mads-Peter V. Christiansen, Bjørk Hammer, Nikolaj Rønne.
# This file is part of AGOX.
# AGOX is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
# AGOX is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. You should
# have received a copy of the GNU General Public License along with AGOX. If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from datetime import datetime
from typing import List, Iterable
from copy import copy

from agox.observer import ObserverHandler, FinalizationHandler
from agox.writer import Writer, get_icon, header_print
from agox.tracker import Tracker
from agox.observer import Observer
from agox.candidates.standard import StandardCandidate
import atexit


class AGOX(ObserverHandler, FinalizationHandler, Writer):
    """
    AGO-X
    Atomistic Global Optimization X
    """

    name = "AGOX"

    def __init__(
        self,
        *args,
        seed=None,
        tracker_kwargs=None,
        candidate_instanstiator=None,
        attach=True,
        report=True,
    ):
        """
        Observers are supplied through *args.
        """
        ObserverHandler.__init__(
            self, handler_identifier="AGOX", dispatch_method=self.run
        )
        FinalizationHandler.__init__(self)
        Writer.__init__(self, verbose=True, use_counter=False, prefix="")

        self._observers = list(args)
        self.candidate_instanstiator = candidate_instanstiator or StandardCandidate
        self.tracker_kwargs = tracker_kwargs or {}

        if seed is not None:
            self.set_seed(seed)

        # This attaches all Observers
        if attach:
            self.update()

        if report:
            self.report()



    def report(self):
        icon = get_icon()
        print(icon)
        header_print("INIT START: AGOX initialization started")
        self.print_observers(hide_log=True)
        self.observer_reports(hide_log=True)
        header_print("INIT FIN: AGOX initialization finished")

    def update(self):
        """
        Calls 'attach' on all Observer-objects in 'self.elements' and updates
        the candidate_instanstiator on all Observer-objects in 'self.elements'
        """
        for observer in self._observers:
            if hasattr(observer, "attach"):
                observer.attach(self)
            if hasattr(observer, "set_candidate_instanstiator"):
                observer.set_candidate_instanstiator(self.candidate_instanstiator)

        # Tracker:
        self.tracker = Tracker(self, **self.tracker_kwargs)


    def set_seed(self, seed):
        if seed is not None:
            np.random.seed(seed)
            self.writer("Numpy random seed: {}".format(seed))

    def run(self, N_iterations, verbose=True, hide_log=True):
        """
        Function called by runscripts that starts the actual optimization procedure.

        This function is controlled by modules attaching themselves as observers to this module.
        The order system ensures that modules with lower order are executed first, but no gurantee within each order,
        so if two modules attach themselves both with order 0 then their individual execution order is not guranteed.
        However, an observer with order 0 will always be executed before an observer with order 1.

        The default ordering system is:
        order = 0: Execution order

        All modules that intend to attach themselves as observers MUST take the order as an argument (with a default value(s)),
        so that if a different order is wanted that can be controlled from runscripts. Do NOT change order default values!
        """

        # Main-loop calling the relevant observers.
        state = State()

        # Force print if something crashes the run.
        atexit.register(self.on_failure)

        self.header_print("RUN START: AGOX run started")

        while (
            state.get_iteration_counter() <= N_iterations
            and not state.get_convergence_status()
        ):
            print("\n")
            self.header_print("Iteration: {}".format(state.get_iteration_counter()))
            dt = datetime.now()
            self.writer("Time: {}".format(dt.strftime("%H:%M:%S")))
            self.writer("Date: {}".format(dt.strftime("%d/%m/%Y")))
            for observer in self.get_observers_in_execution_order():
                observer(state)

            self.header_print("Iteration finished")
            state.clear()
            state.advance_iteration_counter()

        # Some things may want to perform some operation only at the end of the run.
        for method in self.get_finalization_methods():
            method()

        # Remove the force print if everything went well.
        atexit.unregister(self.on_failure)

        self.header_print("RUN FIN: AGOX run finished")

    def add_observer(self, observer: Observer):
        """
        Add an observer to the AGOX object.

        Parameters
        ----------
        observer : Observer
            An observer to add to AGOX.
        """

        if not isinstance(observer, Observer):
            raise TypeError(
                f"Observer must be an instance of Observer: {observer.__name__}"
            )

        self._observers.append(observer)

    def add_observers(self, *observers: Iterable[Observer], update=True, report=True):
        """
        Add a list of modules to the AGOX object.

        Parameters
        ----------
        modules : list
            A list of modules to be added to the AGOX object.
        """

        for observer in observers:
            self.add_observer(observer)

        if update:
            self.update()

        if report:
            self.report()
        
    def on_failure(self):
        """
        This method is called if the the main loop crashes.
        """
        print("RUN FAIL: AGOX run failed.")

class State:
    def __init__(self):
        """
        State object.

        Attributes
        ----------
        cache: dict
            Data communicated between modules is stored in the cache.
        iteration_counter: int
            Keeps track of the number of iterations.
        convergence: bool
            Convergence status, if True the iteration-loop is halted.
        """

        self.cache = {}
        self.iteration_counter = 1
        self.converged = False

    def get_iteration_counter(self):
        """
        Returns
        -------
        int
            The current iteration number.
        """
        return self.iteration_counter

    def set_iteration_counter(self, count):
        """_summary_

        Parameters
        ----------
        count : int
            Iteration count
        """
        self.iteration_counter = count

    def advance_iteration_counter(self):
        """
        Adds one to the iteration counter.
        """
        self.iteration_counter += 1

    def get_from_cache(self, observer, key):
        """

        Gets from the cache with the given key. The observed is passed along
        aswell in order to ensure that the observer is allowed to get with
        that key.

        Parameters
        ----------
        observer : class object
            An AGOX Observer object, e.g. an instance of a Sampler.
        key : str
            The key with which to get something from the cache.

        Returns
        -------
        list
            List of things stored with the given key.
        """
        # Makes sure the module has said it wants to get with this key.
        assert key in observer.get_values
        return copy(self.cache.get(key))

    def add_to_cache(self, observer, key, data, mode):
        """
        Add data to the cache.

        Parameters
        ----------
        observer : class object
            An AGOX Observer object, e.g. an instance of a Sampler.
        key : str
            The key with which to get something from the cache.
        data : list
            List of data to store in the cache.
        mode : str
            Determines the mode in which the data is added to the cache:
            w: Will overwrite existing data with the same key.
            a: Will append to existing data (if there is existing data).
        """
        assert type(data) == list
        # Makes sure the module has said it wants to get with this key.
        assert key in observer.set_values
        assert mode in ["w", "a"]

        if key in self.cache.keys() and mode != "w":
            self.cache[key] += data
        else:
            self.cache[key] = data

    def clear(self):
        """
        Clears the current cachce. Called at the end of each iteration.
        """
        self.cache = {}

    def get_convergence_status(self):
        """
        Returns the convergence status.

        Returns
        -------
        bool
            If True convergence has been reached and the main iteration-loop
            will halt.
        """
        return self.converged

    def set_convergence_status(self, state):
        """
        Set the convergence status.

        Parameters
        ----------
        state : bool
            If True convergence has been reached and the main iteration-loop
            will halt.
        """
        self.converged = state



from typing import Optional
from warnings import warn

import numpy as np
from sklearn.cluster import MiniBatchKMeans

from agox.utils.sparsifiers.ABC_sparsifier import SparsifierBaseClass


class MBkmeans(SparsifierBaseClass):
    name = "MBkmeans"

    def __init__(
        self,
        batch_size: int = 1024,
        fast: bool = False,
        exact_points: bool = False,
        seed: Optional[int] = None,
        **kwargs
    ):
        super().__init__(**kwargs)

        warn(
            "MBkmeans cannot be used with other sparsifiers. "
            "Or together with a method that requires indices of selected features"
            "Use k-medoids instead."
        )

        self.batch_size = batch_size
        self.fast = fast
        self.exact_points = exact_points

        self.cluster = MiniBatchKMeans(
            n_clusters=self.m_points,
            batch_size=self.batch_size,
            random_state=seed,
            init="k-means++",
            n_init=3,
        )

    def sparsify(self, X: np.ndarray) -> np.ndarray:
        if X.shape[0] < self.m_points:
            m_indices = np.arange(0, X.shape[0])
            return X, m_indices

        if hasattr(self.cluster, "cluster_centers_") and self.fast:
            self._MB_episode(X)
        else:
            self.cluster.fit(X)

        if self.exact_points:
            dists = self.cluster.transform(X)
            min_idx = np.argmin(dists, axis=0)
            Xm = X[min_idx, :]
            indices = min_idx
        else:
            Xm = self.cluster.cluster_centers_
            indices = None

        return Xm, indices

    def _MB_episode(self, X: np.ndarray) -> None:
        n_samples = X.shape[0]
        batch_size = min(self.batch_size, n_samples)
        steps = n_samples // batch_size
        random_state = np.random.RandomState()
        for _ in range(steps):
            minibatch_indices = random_state.randint(0, n_samples, batch_size)
            self.cluster.partial_fit(X[minibatch_indices])



import numpy as np
from agox.models.descriptors import Fingerprint
from agox.models.descriptors import SOAP

from ase import Atoms
from agox.environments import Environment

# using AGOX environment
environment = Environment(Atoms(cell=np.eye(3)*10), symbols='H6C6', print_report=False)
global_descriptor = Fingerprint(environment=environment, rc1=5.0)

# using ASE atoms object
atoms = Atoms('H6C6', cell=np.eye(3)*10)
global_descriptor = Fingerprint.from_atoms(atoms, rc1=5.0) 

#using AGOX environment
environment = Environment(Atoms(), symbols='HC', print_report=False)
local_descriptor = SOAP(environment=environment, r_cut=5.0)

#using species list
local_descriptor = SOAP.from_species(['H', 'C'], r_cut=5.0)





import numpy as np
from abc import ABC, abstractmethod
from agox.module import Module, register_modules
from agox.observer import Observer
from agox.writer import Writer, agox_writer

class CollectorBaseClass(ABC, Observer, Writer):

    """
    Base class for all collectors.

    Collectors produce several candidates using a set of generators. 

    Parameters
    ----------
    generators : list
        List of AGOX generator instances.
    sampler : Sampler instance
        An instance of an AGOX sampler
    environment : Environment object.
        An instance of an AGOX environment object.
    order : int, optional
        Order of the collector, by default 2
    sets : dict, optional
        Dictionary of set keys, e.g. {'set_key':'candidates'}. Used to select 
        in which entry in the agox.main.State cache the collector should store the candidates.
    gets : dict, optional
        Dictionary of get keys, e.g. {'get_key':'candidates'}. Used to select 
        from which entry in the agox.main.State cache the collector should get candidates.
    verbose : bool, optional
        If True, print information to the console, by default True.
    use_counter : bool, optional
        If True the writer will not print until the Observer has finished all its 
        action in a given iteration.
    prefix : str, optional
        Prefix for the writer, by default ''.
    surname : str, optional
        Surname for the writer, by default ''.
    """

    def __init__(self, generators=None, sampler=None, environment=None, order=2,
        sets={'set_key':'candidates'}, gets={}, verbose=True, use_counter=True,
        prefix='', surname=''):
        """
        Quite simple, except for lines 50-54 that remove generators as observers,
        as they are now part of a Collector. This may now yield wanted behaviour 
        for generators that need observers_methods that are not just for 
        generation, if that becomes the case an update to those lines are 
        probably necesary. 

        Parameters
        ----------
        generators : list
            List of AGOX generator instances. 
        sampler : Sampler instance
            An instance of an AGOX sampler
        environment : Environment object.
            An instance of an AGOX environment object. 
        """
        Observer.__init__(self, sets=sets, gets=gets, order=order, surname=surname)
        Writer.__init__(self, verbose=verbose, use_counter=use_counter, prefix=prefix)

        assert generators is not None
        assert environment is not None

        self.generators = generators
        self.sampler = sampler
        self.environment = environment
        self.candidates = []
        self.plot_confinement()

        self.add_observer_method(self.generate_candidates,
                                 sets=self.sets[0], gets=self.gets[0], order=self.order[0],
                                 handler_identifier='AGOX')

        for generator in self.generators:            
            observer_methods = [observer_method for observer_method in generator.observer_methods.values()]
            for observer_method in observer_methods:
                if observer_method.method_name == 'generate':
                    generator.remove_observer_method(observer_method)

        # Register the generators as attributes of the collector, so that they can be tracked.
        register_modules(self, self.generators, 'generator')

    @abstractmethod
    def make_candidates(self):
        """
        Method that calls generators to produce a collection of candidates. 

        Returns
        -------
        list 
           List of candidate objects. 
        """
        return candidates
    
    @abstractmethod
    def get_number_of_candidates(self, iteration):
        """
        Method  that computes the number of candidates for the given iteration.

        Parameters
        ----------
        iteration : int
            Iteration count.

        Returns
        -------
        list
            List of integers corresponding to the self.generators list dictating 
            how many candidates are generated (or really how many times each 
            generator is called.)
        """
        # Number of candidates to generate.
        return list_of_integers

    @agox_writer
    @Observer.observer_method
    def generate_candidates(self, state):
        # Make the candidates - this is the method that differs between versions of the class.
        if self.do_check():
            candidates = self.make_candidates()

        iteration = self.get_iteration_counter()
        for candidate in candidates:
            candidate.add_meta_information('search_iteration', iteration)

        # Add to the iteration_cache:
        state.add_to_cache(self, self.set_key, candidates, 'a')
        self.writer('Number of candidates this iteration: {}'.format(len(candidates)))

########################################################################################################################
# Other methods
########################################################################################################################

    def attach(self, main):
        super().attach(main)
        for generator in self.generators:
            generator.attach(main)

    def plot_confinement(self):
        for generator in self.generators: 
            generator.plot_confinement(self.environment)

    def get_fallback_generator(self):
        """
        When sampler is empty but initialized, use a RandomGenerator
        to make new candidates.
        """
        from agox.generators import RandomGenerator
        fallback_generator = RandomGenerator(**self.environment.get_confinement())
        return fallback_generator


def add_example_parser(subparsers):

    parser = subparsers.add_parser('example', help='Example of a CLI tool.')
    parser.set_defaults(func=cli_example)

    parser.add_argument('input', help='Example of positional argument')
    parser.add_argument('-o', '--optional', help='Example of optional argument', default='default value')

def cli_example(args):

    print('Input:', args.input)
    print('Optional:', args.optional)
    print('This is an example CLI')
    return 0





import pytest
from agox.test.run_tests.run_utils import agox_test_run

def test_schnet_bh(tmp_path, cmd_options):
    pytest.skip('This test is not designed to run on the CI')
    mode = 'schnet_bh' # This determines the script that is imported.    
    options = {'tolerance':{'rtol':1e-03, 'atol':1e-03}, 
               'create_mode':cmd_options['create_mode']}
    agox_test_run(mode, tmp_path, options)
